;	ARM7 eForth version 5.06
;	Chen-Hanson Ting,  April 2008

;	Subroutine Threaded Forth Model
;	Adapted to Atmel AT91SAM7X256
;	Assembled under AS in Keil uVision 3
;	Verified on Olimex SAM7-EX256 Board

;	Version 4.03
;	Direct Threaded Forth Model
;	Derived from 80386 eForth versin 4.02
;	and Chien-ja Wu's ARM7 eForth version 1.01

;	Version 5.01
;	Subroutine thread (Branch-Link) model
;	    Register assignments
;		IP	.reg 	R0 	;not used in this model
;		SP	.reg 	R1 
;		RP	.reg 	R2 
;		UP	.reg 	R3 
;		WP	.reg 	R4 
;		TOS	.reg 	R5 
;		XP	.reg 	R6 
;		YP	.reg 	R7 
;	    All Forth words are called by 
;		BL	addr
;	    All low level code words are terminaled by
;		BX	LR
;	    All high level Forth words start with
;		STRFD	RP!,{LR}
;	    All high level Forth words end with
;		LDRFD	RP!,{PC}
;	    Top of data stack is cached in R5
;	UART at 115200 baud, 8 data bits, 1 stop bit, no parity
;	Assembler documentation:
;	    ARM7DTMI Data Sheet
;	    Redhat GNUPro assembler "as"

;	Version 5.02, 09oct04cht
;	Add ADuC702x MMR registers
;	Version 5.03, 10oct04cht
;	Version 5.04, 17may05cht for TingARM
;	Version 5.05, 25aug05cht add SPI interface
;	Version 5.06, 11nov05cht align to gamegirlADuC
;	Version 6.01, 10apr08cht align to at91sam7x256
;		Running under uVision3 RealView from Keil
;		Change PC relative jumps to branches in UMUL, CMOVE, and MOVE
;		Add CONSTANT, Modify VARIABLE and CREATE
;	Version 6.02, 05may08cht
;		Running in RAM
;		Change BL BRAN to B 
;	Version 6.03, 12AUG08cht
;		CP initiated at COLD2
;		Flash programmable 
;		User area in RAM, no need to move
;	Version 6.04 26SEP08cht
;		Change baud rate back to 9600

        AREA	EFORTH, CODE
        ARM
        EXPORT COLD

; Version control

VER EQU    0x06			;major release version
EXT EQU    0x04			;minor extension

; Constants

UARTBASE EQU	0xFFFF0700
ROMBASE  EQU	0x00100000
RAMBASE  EQU	0x00200000
;RAMOFFSET  EQU	0x00000000			;absolute
;MAPOFFSET  EQU	0x00000000		 	;absolute
RAMOFFSET  EQU	0x00200000	;remap
MAPOFFSET  EQU	0x00100000	;remap

COMPO EQU    0x040			;lexicon compile only 
IMEDD EQU    0x080			;lexicon immediate bit
MASKK EQU    0x07F7F7F1F		;lexicon bit mask

CELLL EQU    4				;size of a cell
BASEE EQU    10			;default radix
VOCSS EQU    8				;depth of vocabulary stack

BKSPP EQU    8				;backspace
LF EQU     10				;line feed
CRR EQU    13				;carriage return
ERR EQU    27				;error escape
TIC EQU    39				;tick

CALLL EQU    0x0EB000000		;BL opcodes

; GPIO (PIOA) definitions
PIOA_BASE       	equ     0xFFFFF400	; PIOA Base Address
PIO_PDR        equ     0x04           ; PIOA
PIO_IDR        equ     0x44           ; PIOA	interrupt disable
PIO_ASR        equ     0x70           ; PIOA
PIO_MASK        equ     0x18000000    ; PIOA

; Debug Controller (DBGU) definitions
DBGU_BASE       	equ     0xFFFFF200   ; DBGU Base Address
DBGU_CR        equ     0x00           ; DBGU
DBGU_MR        equ     0x04           ; DBGU	 
DBGU_IER        equ     0x08          ; DBGU
DBGU_IDR        equ     0x0C      	; DBGU  
DBGU_IMR        equ     0x10      	; DBGU  
DBGU_SR        equ     0x14      		; DBGU  
DBGU_RHR        equ     0x18      	; DBGU  
DBGU_THR        equ     0x1C      	; DBGU  
DBGU_BRGR       	equ     0x20      	; DBGU  
DBGU_PTCR       	equ     0x120      	; DBGU  

; Memory allocation    0//code>--//--<name//up>--<sp//tib>--rp//em
;	000000				;mapped memory
;	100000				;ram
;	101000				;inital DP
;	10fc00				;top of data stack
;	10fc00				;TIB terminal input buffer
;	10fd00				;top of return stack
;	10fd00				;system variables
;	10ffff				;end of ram
;	200000				;flash
;	200300				;start of eForth
;	23ffff				;end of flash

SPP EQU    0x20FC00-RAMOFFSET			;top of data stack (SP0)
TIBB EQU   0x20FC00-RAMOFFSET			;terminal input buffer (TIB)
RPP EQU    0x20FD00-RAMOFFSET			;top of return stack (RP0)
UPP EQU    0x20FD00-RAMOFFSET			;start of user area (UP0)

; Initialize assembly variables

;       Assemble inline direct threaded code ending.

 	MACRO	
 	_NEXT   			;end low level word
	BX	LR
	MEND

 	MACRO	
 	_NEST   			;start high level word
	STMFD	R2!,{LR}
	MEND

 	MACRO	
 	_UNNEST   			;end high level word
	LDMFD	R2!,{PC}
	MEND

 	MACRO	
 	_DOVAR   			;variables
	STR	R5,[R1,#-4]!
	MOV	R5,PC
	BX	LR
	MEND

 	MACRO	
 	_DOLIT   			;long literals
	BL	DOLIT
	MEND

 	MACRO	
 	_PUSH   			;push R5 on data stack
	STR	R5,[R1,#-4]!
	MEND

 	MACRO	
 	_POP   				;pop data stack to R5
	LDR	R5,[R1],#4
	MEND

; COLD start moves the following to USER variables.
; MUST BE IN SAME ORDER AS USER VARIABLES.

	ALIGN 256  			;align to page boundary

UZERO          DCD      0,0,0,0	;reserved
		DCD    SPP  		;SP0
		DCD    RPP  		;RP0
		DCD    QRX-MAPOFFSET  		;'?KEY
		DCD    TXSTO-MAPOFFSET  	;'EMIT
		DCD    ACCEP-MAPOFFSET  	;'EXPECT
		DCD    KTAP-MAPOFFSET  		;'TAP
		DCD    TXSTO-MAPOFFSET  	;'ECHO
		DCD    DOTOK-MAPOFFSET  	;'PROMPT
		DCD    BASEE  	;BASE
		DCD      0		;tmp
		DCD      0		;SPAN
		DCD      0		;>IN
		DCD      0		;#TIB
		DCD      TIBB		;TIB
		DCD      0		;CSP
		DCD      INTER-MAPOFFSET	;'EVAL
		DCD      NUMBQ-MAPOFFSET	;'NUMBER
		DCD      0		;HLD
		DCD      0		;HANDLER
		DCD      0		;CONTEXT pointer
		DCD      0,0,0,0,0,0,0,0	;vocabulary stack
		DCD      0		;CURRENT pointer
		DCD      0		;vocabulary link pointer
		DCD      COLD2-MAPOFFSET	;CP
		DCD      0		;NP
		DCD      LASTN-MAPOFFSET	;LAST
		DCD      LASTN-MAPOFFSET	;FORTH dictionary
ULAST

		ALIGN	
;   ?RX         ( -- c T | F )
;               Return input character and true, or a false if no input.
		DCD	0
_QRX		DCB   3
		DCB	"?RX"
		ALIGN
QRX
        MVN R4, #0x0E00			   ;FFFFF1FF
 		LDR	R6, [R4, #DBGU_SR+1]  ;dbgu_sr
 		ANDS	R6, R6, #1
		BEQ     QRX1
		STR	R5,[R1,#-4]!
 		LDR	R5, [R4, #DBGU_RHR+1]
		STR	R5,[R1,#-4]!
		MVN	R5,#0
		_NEXT
QRX1		STR	R5,[R1,#-4]!
		MOV	R5,#0
		_NEXT

;   TX!         ( c -- )
;               Send character c to the output device.

		DCD	_QRX-MAPOFFSET
_TXSTO		DCB   3
		DCB	"TX!"
		ALIGN	
TXSTO
TX2
        MVN R4, #0x0E00				;FFFFF1FF
 		LDR	R6, [R4, #DBGU_SR+1]  ;dbgu_sr
 		ANDS	R6, R6, #0X200
 		BEQ	TX2
 		STR	R5, [R4, #DBGU_THR+1] ;dbgu_thr
		LDR	R5,[R1],#4
		_NEXT

;   !IO         ( -- )
;               Initialize the serial I/O devices.

		DCD	_TXSTO-MAPOFFSET
_STOIO		DCB   3
		DCB	"!IO"
		ALIGN	
STOIO
		MVN R4, #0x0C00				 ;FFFFF3FF
        MOV R6, #0x18000000
 		STR	R6, [R4, #PIO_IDR+1]	;pio_idr
 		STR	R6, [R4, #PIO_ASR+1]	;pio_asr
 		STR	R6, [R4, #PIO_PDR+1]	;pio_pdr
		MVN R4, #0x0F00				;FFFFF0FF
        MVN R6, #0
 		STR	R6, [R4, #0x25]			;aic_idcr
 		STR	R6, [R4, #0x29]			;aic_iccr
        MVN R4, #0x0E00				;FFFFF1FF
 		MOV	R6, #0X0AC   			;reset TX RX
 		STR	R6, [R4, #DBGU_CR+1]
 		MVN	R6, #0		 			;disable interrupts
 		STR	R6, [R4, #DBGU_IDR+1]
 		MOV	R6, #0X800  			;mode
 		STR	R6, [R4, #DBGU_MR+1]
; 		MOV	R6, #26 				;brgr 115200 baud
 		MOV	R6, #312 				;brgr 9600 baud
 		STR	R6, [R4, #DBGU_BRGR+1]
 		MOV	R6, #0x200
		ADD R6, R6, #2 			;disable DMA
 		STR	R6, [R4, #DBGU_PTCR+1]
 		MOV	R6, #0X50				;enable TX RX
 		STR	R6, [R4, #DBGU_CR+1]
 		LDR	R6, [R4, #0x41]	  		;dbgu_cidr
 		LDR	R6, [R4, #0x45]			;dbgu_exid
		_NEXT

; The kernel

;   NOP       ( -- w )
;               Push an inline literal.

		DCD	_STOIO-MAPOFFSET
_NOP		DCB   3
		DCB	"NOP"
		ALIGN	
NOP
		_NEXT

;   doLIT       ( -- w )
;               Push an inline literal.

		DCD	_NOP-MAPOFFSET
_LIT		DCB   COMPO+5
		DCB	"doLIT"
		ALIGN	
DOLIT
		_PUSH
		LDR	R5,[LR],#4
		_NEXT

;   EXECUTE     ( ca -- )
;               Execute the word at ca.

		DCD	_LIT-MAPOFFSET
_EXECU		DCB   7
		DCB	"EXECUTE"
		ALIGN	
EXECU
		MOV	R4,R5
		_POP
		BX	R4

;   next        ( -- )
;               Run time code for the single index loop.
;               : next ( -- ) \ hilevel model
;                 r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

		DCD	_EXECU-MAPOFFSET
_DONXT		DCB   COMPO+4
		DCB	"next"
		ALIGN	
DONXT
		LDR	R4,[R2]
		MOVS	R4,R4
		BNE	NEXT1
		ADD	R2,R2,#4
		ADD	LR,LR,#4
		_NEXT
NEXT1		SUB	R4,R4,#1
		STR	R4,[R2]
		LDR	LR,[LR]
		_NEXT

;   ?branch     ( f -- )
;               Branch if flag is zero.

		DCD	_DONXT-MAPOFFSET
_QBRAN		DCB   COMPO+7
		DCB	"?branch"
		ALIGN	
QBRAN
		MOVS	R4,R5
		_POP
		BNE	QBRAN1
		LDR	LR,[LR]
		_NEXT
QBRAN1		ADD	LR,LR,#4
		_NEXT

;   branch      ( -- )
;               Branch to an inline address.

		DCD	_QBRAN-MAPOFFSET
_BRAN		DCB   COMPO+6
		DCB	"branch"
		ALIGN	
BRAN
		LDR	LR,[LR]
		_NEXT

;   !           ( w a -- )
;               Pop the data stack to memory.

		DCD	_BRAN-MAPOFFSET
_STORE		DCB   1
		DCB	"!"
		ALIGN	
STORE
		LDR	R4,[R1],#4
		STR	R4,[R5]
		_POP
		_NEXT

;   @           ( a -- w )
;               Push memory location to the data stack.

		DCD	_STORE-MAPOFFSET
_AT		DCB   1
		DCB	"@"
		ALIGN	
AT
		LDR	R5,[R5]
		_NEXT

;   W!           ( w a -- )
;               Pop the data stack to memory.

		DCD	_AT-MAPOFFSET
_WSTOR		DCB   2
		DCB	"W!"
		ALIGN	
WSTOR
		LDR	R4,[R1],#4
		STRH	R4,[R5],#0
		_POP
		_NEXT

;   W@           ( a -- w )
;               Push memory location to the data stack.

		DCD	_WSTOR-MAPOFFSET
_WAT		DCB   2
		DCB	"W@"
		ALIGN	
WAT
		LDRH	R5,[R5]
		_NEXT

;   C!          ( c b -- )
;               Pop the data stack to byte memory.

		DCD	_WAT-MAPOFFSET
_CSTOR		DCB   2
		DCB	"C!"
		ALIGN	
CSTOR
		LDR	R4,[R1],#4
		STRB	R4,[R5]
		_POP
		_NEXT

;   C@          ( b -- c )
;               Push byte memory location to the data stack.

		DCD	_CSTOR-MAPOFFSET
_CAT		DCB   2
		DCB	"C@"
		ALIGN	
CAT
		LDRB	R5,[R5]
		_NEXT

;   RP@         ( -- a )
;               Push the current RP to the data stack.

		DCD	_CAT-MAPOFFSET
_RPAT		DCB   3
		DCB	"RP@"
		ALIGN	
RPAT
		_PUSH
		MOV	R5,R2
		_NEXT

;   RP!         ( a -- )
;               Set the return stack pointer.

		DCD	_RPAT-MAPOFFSET
_RPSTO		DCB   COMPO+3
		DCB	"RP!"
		ALIGN	
RPSTO
		MOV	R2,R5
		_POP
		_NEXT

;   R>          ( -- w )
;               Pop the return stack to the data stack.

		DCD	_RPSTO-MAPOFFSET
_RFROM		DCB   2
		DCB	"R>"
		ALIGN	
RFROM
		_PUSH
		LDR	R5,[R2],#4
		_NEXT

;   R@          ( -- w )
;               Copy top of return stack to the data stack.

		DCD	_RFROM-MAPOFFSET
_RAT		DCB   2
		DCB	"R@"
		ALIGN	
RAT
		_PUSH
		LDR	R5,[R2]
		_NEXT

;   >R          ( w -- )
;               Push the data stack to the return stack.

		DCD	_RAT-MAPOFFSET
_TOR		DCB   COMPO+2
		DCB	">R"
		ALIGN	
TOR
		STR	R5,[R2,#-4]!
		_POP
		_NEXT

;   SP@         ( -- a )
;               Push the current data stack pointer.

		DCD	_TOR-MAPOFFSET
_SPAT		DCB   3
		DCB	"SP@"
		ALIGN	
SPAT
		_PUSH
		MOV	R5,R1
		_NEXT

;   SP!         ( a -- )
;               Set the data stack pointer.

		DCD	_SPAT-MAPOFFSET
_SPSTO		DCB   3
		DCB	"SP!"
		ALIGN	
SPSTO
		MOV	R1,R5
		_NEXT

;   DROP        ( w -- )
;               Discard top stack item.

		DCD	_SPSTO-MAPOFFSET
_DROP		DCB   4
		DCB	"DROP"
		ALIGN	
DROP
		_POP
		_NEXT

;   DUP         ( w -- w w )
;               Duplicate the top stack item.

		DCD	_DROP-MAPOFFSET
_DUPP		DCB   3
		DCB	"DUP"
		ALIGN	
DUPP
		_PUSH
		_NEXT

;   SWAP        ( w1 w2 -- w2 w1 )
;               Exchange top two stack items.

		DCD	_DUPP-MAPOFFSET
_SWAP		DCB   4
		DCB	"SWAP"
		ALIGN	
SWAP
		LDR	R4,[R1]
		STR	R5,[R1]
		MOV	R5,R4
		_NEXT

;   OVER        ( w1 w2 -- w1 w2 w1 )
;               Copy second stack item to top.

		DCD	_SWAP-MAPOFFSET
_OVER		DCB   4
		DCB	"OVER"
		ALIGN	
OVER
		_PUSH
		LDR	R5,[R1,#4]
		_NEXT

;   0<          ( n -- t )
;               Return true if n is negative.

		DCD	_OVER-MAPOFFSET
_ZLESS		DCB   2
		DCB	"0<"
		ALIGN	
ZLESS
		MOV	R4,#0
		ADD	R5,R4,R5,ASR #32
		_NEXT

;   AND         ( w w -- w )
;               Bitwise AND.

		DCD	_ZLESS-MAPOFFSET
_ANDD		DCB   3
		DCB	"AND"
		ALIGN	
ANDD
		LDR	R4,[R1],#4
		AND     R5,R5,R4
		_NEXT

;   OR          ( w w -- w )
;               Bitwise inclusive OR.

		DCD	_ANDD-MAPOFFSET
_ORR		DCB   2
		DCB	"OR"
		ALIGN	
ORR
		LDR	R4,[R1],#4
		ORR     R5,R5,R4
		_NEXT

;   XOR         ( w w -- w )
;               Bitwise exclusive OR.

		DCD	_ORR-MAPOFFSET
_XORR		DCB   3
		DCB	"XOR"
		ALIGN	
XORR
		LDR	R4,[R1],#4
		EOR     R5,R5,R4
		_NEXT

;   RSHIFT         ( w # -- w )
;               Right shift # bits.

		DCD	_XORR-MAPOFFSET
_RSHIFT		DCB   6
		DCB	"RSHIFT"
		ALIGN	
RSHIFT
		LDR	R4,[R1],#4
		MOV     R5,R4,ASR R5
		_NEXT

;   LSHIFT         ( w # -- w )
;               Right shift # bits.

		DCD	_RSHIFT-MAPOFFSET
_LSHIFT		DCB   6
		DCB	"LSHIFT"
		ALIGN	
LSHIFT
		LDR	R4,[R1],#4
		MOV     R5,R4,LSL R5
		_NEXT

;   UM+         ( w w -- w cy )
;               Add two numbers, return the sum and carry flag.

		DCD	_LSHIFT-MAPOFFSET
_UPLUS		DCB   3
		DCB	"UM+"
		ALIGN	
UPLUS
		LDR	R4,[R1]
		ADDS	R4,R4,R5
		MOV	R5,#0
		ADC	R5,R5,#0
		STR	R4,[R1]
		_NEXT

;   +         ( w w -- w )
;               Add.

		DCD	_UPLUS-MAPOFFSET
_PLUS		DCB   1
		DCB	"+"
		ALIGN	
PLUS
		LDR	R4,[R1],#4
		ADD     R5,R5,R4
		_NEXT

;   -         ( w w -- w )
;               Subtract.

		DCD	_PLUS-MAPOFFSET
_SUBB		DCB   1
		DCB	"-"
		ALIGN	
SUBB
		LDR	R4,[R1],#4
		RSB     R5,R5,R4
		_NEXT

;   *         ( w w -- w )
;               Multiply.

		DCD	_SUBB-MAPOFFSET
_STAR		DCB   1
		DCB	"*"
		ALIGN	
STAR
		LDR	R4,[R1],#4
		MUL     R5,R4,R5
		_NEXT

;   UM*         ( w w -- ud )
;               Unsigned multiply.

		DCD	_STAR-MAPOFFSET
_UMSTA		DCB   3
		DCB	"UM*"
		ALIGN	
UMSTA
		LDR	R4,[R1]
		UMULL	R6,R7,R5,R4
		STR	R6,[R1]
		MOV	R5,R7
		_NEXT

;   M*         ( w w -- d )
;               Unsigned multiply.

		DCD	_UMSTA-MAPOFFSET
_MSTAR		DCB   2
		DCB	"M*"
		ALIGN	
MSTAR
		LDR	R4,[R1]
		SMULL	R6,R7,R5,R4
		STR	R6,[R1]
		MOV	R5,R7
		_NEXT

;   1+         ( w -- w+1 )
;               Add 1.

		DCD	_MSTAR-MAPOFFSET
_ONEP		DCB   2
		DCB	"1+"
		ALIGN	
ONEP
		ADD     R5,R5,#1
		_NEXT

;   1-         ( w -- w-1 )
;               Subtract 1.

		DCD	_ONEP-MAPOFFSET
_ONEM		DCB   2
		DCB	"1-"
		ALIGN	
ONEM
		SUB     R5,R5,#1
		_NEXT

;   2+         ( w -- w+2 )
;               Add 1.

		DCD	_ONEM-MAPOFFSET
_TWOP		DCB   2
		DCB	"2+"
		ALIGN	
TWOP
		ADD     R5,R5,#2
		_NEXT

;   2-         ( w -- w-2 )
;               Subtract 2.

		DCD	_TWOP-MAPOFFSET
_TWOM		DCB   2
		DCB	"2-"
		ALIGN	
TWOM
		SUB     R5,R5,#2
		_NEXT

;   CELL+       ( w -- w+4 )
;               Add 4.

		DCD	_TWOM-MAPOFFSET
_CELLP		DCB   5
		DCB	"CELL+"
		ALIGN	
CELLP
		ADD     R5,R5,#4
		_NEXT

;   CELL-       ( w -- w-4 )
;               Subtract 4.

		DCD	_CELLP-MAPOFFSET
_CELLM		DCB   5
		DCB	"CELL-"
		ALIGN	
CELLM
		SUB     R5,R5,#4
		_NEXT

;   BL       	( -- 32 )
;               Blank (ASCII space).

		DCD	_CELLM-MAPOFFSET
_BLANK		DCB   2
		DCB	"BL"
		ALIGN	
BLANK
		_PUSH
		MOV	R5,#32
		_NEXT

;   CELLS       ( w -- w*4 )
;               Multiply 4.

		DCD	_BLANK-MAPOFFSET
_CELLS		DCB   5
		DCB	"CELLS"
		ALIGN	
CELLS
		MOV	R5,R5,LSL#2
		_NEXT

;   CELL/       ( w -- w*4 )
;               Divide by 4.

		DCD	_CELLS-MAPOFFSET
_CELLSL		DCB   5
		DCB	"CELL/"
		ALIGN	
CELLSL
		MOV	R5,R5,ASR#2
		_NEXT

;   2*       ( w -- w*2 )
;               Multiply 2.

		DCD	_CELLSL-MAPOFFSET
_TWOST		DCB   2
		DCB	"2*"
		ALIGN	
TWOST
		MOV	R5,R5,LSL#1
		_NEXT

;   2/       ( w -- w/2 )
;               Divide by 2.

		DCD	_TWOST-MAPOFFSET
_TWOSL		DCB   2
		DCB	"2/"
		ALIGN	
TWOSL
		MOV	R5,R5,ASR#1
		_NEXT

;   8LSHIFT     ( w -- w<<8 )
;               Shift left 8 bits.

		DCD	_TWOSL-MAPOFFSET
_LSH8		DCB   7
		DCB	"8LSHIFT"
		ALIGN	
LSH8
		MOV	R5,R5,LSL#8
		_NEXT

;   16LSHIFT     ( w -- w<<16 )
;               Shift left 16 bits.

		DCD	_LSH8-MAPOFFSET
_LSH16		DCB   8
		DCB	"16LSHIFT"
		ALIGN	
LSH16
		MOV	R5,R5,LSL#16
		_NEXT

;   24LSHIFT     ( w -- w<<24 )
;               Shift left 24 bits.

		DCD	_LSH16-MAPOFFSET
_LSH24		DCB   8
		DCB	"24LSHIFT"
		ALIGN	
LSH24
		MOV	R5,R5,LSL#24
		_NEXT

;   8RSHIFT     ( w -- w>>8 )
;               Shift right 8 bits.

		DCD	_LSH24-MAPOFFSET
_RSH8		DCB   7
		DCB	"8RSHIFT"
		ALIGN	
RSH8
		MOV	R5,R5,LSR#8
		_NEXT

;   16RSHIFT     ( w -- w>>16 )
;               Shift right 16 bits.

		DCD	_RSH8-MAPOFFSET
_RSH16		DCB   8
		DCB	"16RSHIFT"
		ALIGN	
RSH16
		MOV	R5,R5,LSR#16
		_NEXT

;   24RSHIFT     ( w -- w>>24 )
;               Shift right 24 bits.

		DCD	_RSH16-MAPOFFSET
_RSH24		DCB   8
		DCB	"24RSHIFT"
		ALIGN	
RSH24
		MOV	R5,R5,LSR#24
		_NEXT

;   ?DUP	( w -- w w | 0 )
;               Conditional duplicate.

		DCD	_RSH24-MAPOFFSET
_QDUP		DCB   4
		DCB	"?DUP"
		ALIGN	
QDUP
		MOVS	R4,R5
		STRNE	R5,[R1,#-4]!
		_NEXT

;   ROT		( w1 w2 w3 -- w2 w3 w1 )
;               Rotate top 3 items.

		DCD	_QDUP-MAPOFFSET
_ROT		DCB   3
		DCB	"ROT"
		ALIGN	
ROT
		LDR	R4,[R1]
		STR	R5,[R1]
		LDR	R5,[R1,#4]
		STR	R4,[R1,#4]
		_NEXT

;   2DROP	( w1 w2 -- )
;               Drop top 2 items.

		DCD	_ROT-MAPOFFSET
_DDROP		DCB   5
		DCB	"2DROP"
		ALIGN	
DDROP
		_POP
		_POP
		_NEXT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;               Duplicate top 2 items.

		DCD	_DDROP-MAPOFFSET
_DDUP		DCB   4
		DCB	"2DUP"
		ALIGN	
DDUP
		LDR	R4,[R1]
		STR	R5,[R1,#-4]!
		STR	R4,[R1,#-4]!
		_NEXT

;   D+		( d1 d2 -- d3 )
;               Add top 2 double numbers.

		DCD	_DDUP-MAPOFFSET
_DPLUS		DCB   2
		DCB	"D+"
		ALIGN	
DPLUS
		LDR	R4,[R1],#4
		LDR	R6,[R1],#4
		LDR	R7,[R1]
		ADDS	R4,R4,R7
		STR	R4,[R1]
		ADC	R5,R5,R6
		_NEXT

;   NOT         ( w -- !w )
;               1"s complement.

		DCD	_DPLUS-MAPOFFSET
_INVER		DCB   3
		DCB	"NOT"
		ALIGN	
INVER
		MVN     R5,R5
		_NEXT

;   NEGATE	( w -- -w )
;              2's complement.

		DCD	_INVER-MAPOFFSET
_NEGAT		DCB   6
		DCB	"NEGATE"
		ALIGN	
NEGAT
		RSB     R5,R5,#0
		_NEXT

;   ABS         ( w -- |w| )
;               Absolute.

		DCD	_NEGAT-MAPOFFSET
_ABSS		DCB   3
		DCB	"ABS"
		ALIGN	
ABSS
		TST	R5,#0x80000000
		RSBNE   R5,R5,#0
		_NEXT

;   =         ( w w -- t )
;               Equal?

		DCD	_ABSS-MAPOFFSET
_EQUAL		DCB   1
		DCB	"="
		ALIGN	
EQUAL
		LDR	R4,[R1],#4
		CMPS    R5,R4
		MVNEQ	R5,#0
		MOVNE	R5,#0
		_NEXT

;   U<         ( w w -- t )
;               Unsigned equal?

		DCD	_EQUAL-MAPOFFSET
_ULESS		DCB   2
		DCB	"U<"
		ALIGN	
ULESS
		LDR	R4,[R1],#4
		CMPS    R4,R5
		MVNCC	R5,#0
		MOVCS	R5,#0
		_NEXT

;   <		( w w -- t )
;               Less?

		DCD	_ULESS-MAPOFFSET
_LESS		DCB   1
		DCB	"<"
		ALIGN	
LESS
		LDR	R4,[R1],#4
		CMPS    R4,R5
		MVNLT	R5,#0
		MOVGE	R5,#0
		_NEXT

;   MAX         ( w w -- max )
;               Leave maximum.

		DCD	_LESS-MAPOFFSET
_MAX		DCB   3
		DCB	"MAX"
		ALIGN	
MAX
		LDR	R4,[R1],#4
		CMPS    R4,R5
		MOVGT	R5,R4
		_NEXT

;   MIN         ( w w -- min )
;               Leave minimum.

		DCD	_MAX-MAPOFFSET
_MIN		DCB   3
		DCB	"MIN"
		ALIGN	
MIN
		LDR	R4,[R1],#4
		CMPS    R4,R5
		MOVLT	R5,R4
		_NEXT

;   +!         ( w a -- )
;               Add to memory.

		DCD	_MIN-MAPOFFSET
_PSTOR		DCB   2
		DCB	"+!"
		ALIGN	
PSTOR
		LDR	R4,[R1],#4
		LDR	R6,[R5]
		ADD	R6,R6,R4
		STR	R6,[R5]
		_POP
		_NEXT

;   2!         ( d a -- )
;               Store double number.

		DCD	_PSTOR-MAPOFFSET
_DSTOR		DCB   2
		DCB	"2!"
		ALIGN	
DSTOR
		LDR	R4,[R1],#4
		LDR	R6,[R1],#4
		STR	R4,[R5],#4
		STR	R6,[R5]
		_POP
		_NEXT

;   2@         ( a -- d )
;               Fetch double number.

		DCD	_DSTOR-MAPOFFSET
_DAT		DCB   2
		DCB	"2@"
		ALIGN	
DAT
		LDR	R4,[R5,#4]
		STR	R4,[R1,#-4]!
		LDR	R5,[R5]
		_NEXT

;   COUNT       ( b -- b+1 c )
;               Fetch length ofDCB.

		DCD	_DAT-MAPOFFSET
_COUNT		DCB   5
		DCB	"COUNT"
		ALIGN	
COUNT
		LDRB	R4,[R5],#1
		_PUSH
		MOV	R5,R4
		_NEXT

;   DNEGATE     ( d -- -d )
;               Negate double number.

		DCD	_COUNT-MAPOFFSET
_DNEGA		DCB   7
		DCB	"DNEGATE"
		ALIGN	
DNEGA
		LDR	R4,[R1]
		RSBS	R4,R4,#0
		RSC	R5,R5,#0
		STR	R4,[R1]
		_NEXT

; System and user variables

;   doVAR       ( -- a )
;               Run time routine for VARIABLE and CREATE.

		DCD	_DNEGA-MAPOFFSET
_DOVAR		DCB  COMPO+5
		DCB	"doVAR"
		ALIGN	
DOVAR
		_PUSH
		MOV	R5,LR
		_UNNEST



;   doCON       ( -- a ) 
;               Run time r outine for CONSTANT.

		DCD	_DOVAR-MAPOFFSET
_DOCON		DCB  COMPO+5
		DCB	"doCON"
		ALIGN	
DOCON
		_PUSH
		LDR	R5,[LR]
		_UNNEST

;   UP          ( -- a )
;               Pointer to the user area.

		DCD	_DOCON-MAPOFFSET
_UP		DCB  2
		DCB	"UP"
		ALIGN	
UP
		_PUSH
		MOV	R5,R3
		_NEXT

;   doUSER      ( -- a )
;               Run time routine for user variables.

		DCD	_UP-MAPOFFSET
_DOUSE		DCB  COMPO+6
		DCB	"doUSER"
		ALIGN	
DOUSE
		_PUSH
		LDR	R4,[LR]
		ADD	R5,R4,R3
		_UNNEST

;   SP0         ( -- a )
;               Pointer to bottom of the data stack.

		DCD	_DOUSE-MAPOFFSET
_SZERO		DCB   3
		DCB	"SP0"
		ALIGN	
SZERO
;		_PUSH
;		ADD	R5,R3,#16
;		_NEXT
		_NEST
		BL	DOUSE
		DCD	16

;   RP0         ( -- a )
;               Pointer to bottom of the return stack.

		DCD	_SZERO-MAPOFFSET
_RZERO		DCB   3
		DCB	"RP0"
		ALIGN	
RZERO
		_PUSH
		ADD	R5,R3,#20
		_NEXT

;   '?KEY       ( -- a )
;               Execution vector of ?KEY.

		DCD	_RZERO-MAPOFFSET
_TQKEY		DCB   5
		DCB	"'?KEY"
		ALIGN	
TQKEY
		_PUSH
		ADD	R5,R3,#24
		_NEXT

;   'EMIT       ( -- a )
;               Execution vector of EMIT.

		DCD	_TQKEY-MAPOFFSET
_TEMIT		DCB   5
		DCB	"'EMIT"
		ALIGN	
TEMIT
		_PUSH
		ADD	R5,R3,#28
		_NEXT

;   'EXPECT     ( -- a )
;               Execution vector of EXPECT.

		DCD	_TEMIT-MAPOFFSET
_TEXPE		DCB   7
		DCB	"'EXPECT"
		ALIGN	
TEXPE
		_PUSH
		ADD	R5,R3,#32
		_NEXT

;   'TAP        ( -- a )
;               Execution vector of TAP.

		DCD	_TEXPE-MAPOFFSET
_TTAP		DCB   4
		DCB	"'TAP"
		ALIGN	
TTAP
		_PUSH
		ADD	R5,R3,#36
		_NEXT

;   'ECHO       ( -- a )
;               Execution vector of ECHO.

		DCD	_TTAP-MAPOFFSET
_TECHO		DCB   5
		DCB	"'ECHO"
		ALIGN	
TECHO
		_PUSH
		ADD	R5,R3,#40
		_NEXT

;   'PROMPT     ( -- a )
;               Execution vector of PROMPT.

		DCD	_TECHO-MAPOFFSET
_TPROM		DCB   7
		DCB	"'PROMPT"
		ALIGN	2
TPROM
		_PUSH
		ADD	R5,R3,#44
		_NEXT

;   BASE        ( -- a )
;               Storage of the radix base for numeric I/O.

		DCD	_TPROM-MAPOFFSET
_BASE		DCB   4
		DCB	"BASE"
		ALIGN	
BASE
		_PUSH
		ADD	R5,R3,#48
		_NEXT

;   tmp         ( -- a )
;               A temporary storage location used in parse and find.

		DCD	_BASE-MAPOFFSET
_TEMP		DCB   COMPO+3
		DCB	"tmp"
		ALIGN	
TEMP
		_PUSH
		ADD	R5,R3,#52
		_NEXT

;   SPAN        ( -- a )
;               Hold character count received by EXPECT.

		DCD	_TEMP-MAPOFFSET
_SPAN		DCB   4
		DCB	"SPAN"
		ALIGN	
SPAN
		_PUSH
		ADD	R5,R3,#56
		_NEXT

;   >IN         ( -- a )
;               Hold the character pointer while parsing input stream.

		DCD	_SPAN-MAPOFFSET
_INN		DCB   3
		DCB	">IN"
		ALIGN	
INN
		_PUSH
		ADD	R5,R3,#60
		_NEXT

;   #TIB        ( -- a )
;               Hold the current count and address of the terminal input buffer.

		DCD	_INN-MAPOFFSET
_NTIB		DCB   4
		DCB	"#TIB"
		ALIGN	
NTIB
		_PUSH
		ADD	R5,R3,#64
		_NEXT

;   CSP         ( -- a )
;               Hold the stack pointer for error checking.

		DCD	_NTIB-MAPOFFSET
_CSP		DCB   3
		DCB	"CSP"
		ALIGN	
CSP
		_PUSH
		ADD	R5,R3,#72
		_NEXT

;   'EVAL       ( -- a )
;               Execution vector of EVAL.

		DCD	_CSP-MAPOFFSET
_TEVAL		DCB   5
		DCB	"'EVAL"
		ALIGN	
TEVAL
		_PUSH
		ADD	R5,R3,#76
		_NEXT

;   'NUMBER     ( -- a )
;               Execution vector of NUMBER?.

		DCD	_TEVAL-MAPOFFSET
_TNUMB		DCB   7
		DCB	"'NUMBER"
		ALIGN	
TNUMB
		_PUSH
		ADD	R5,R3,#80
		_NEXT

;   HLD         ( -- a )
;               Hold a pointer in building a numeric outputDCB.

		DCD	_TNUMB-MAPOFFSET
_HLD		DCB   3
		DCB	"HLD"
		ALIGN	
HLD
		_PUSH
		ADD	R5,R3,#84
		_NEXT

;   HANDLER     ( -- a )
;               Hold the return stack pointer for error handling.

		DCD	_HLD-MAPOFFSET
_HANDL		DCB   7
		DCB	"HANDLER"
		ALIGN	
HANDL
		_PUSH
		ADD	R5,R3,#88
		_NEXT

;   CONTEXT     ( -- a )
;               A area to specify vocabulary search order.

		DCD	_HANDL-MAPOFFSET
_CNTXT		DCB   7
		DCB	"CONTEXT"
		ALIGN	
CNTXT
		_PUSH
		ADD	R5,R3,#92
		_NEXT

;   CURRENT     ( -- a )
;               Point to the vocabulary to be extended.

		DCD	_CNTXT-MAPOFFSET
_CRRNT		DCB   7
		DCB	"CURRENT"
		ALIGN	
CRRNT
		_PUSH
		ADD	R5,R3,#128
		_NEXT

;   CP          ( -- a )
;               Point to the top of the code dictionary.

		DCD	_CRRNT-MAPOFFSET
_CP		DCB   2
		DCB	"CP"
		ALIGN	
CP
		_PUSH
		ADD	R5,R3,#136
		_NEXT

;   NP          ( -- a )
;               Point to the bottom of the name dictionary.

		DCD	_CP-MAPOFFSET
_NP		DCB   2
		DCB	"NP"
		ALIGN	
NP
		_PUSH
		ADD	R5,R3,#140
		_NEXT

;   LAST        ( -- a )
;               Point to the last name in the name dictionary.

		DCD	_NP-MAPOFFSET
_LAST		DCB   4
		DCB	"LAST"
		ALIGN	
LAST
		_PUSH
		ADD	R5,R3,#144
		_NEXT

;   'FORTH        ( -- a )
;               Point to the last name in the name dictionary.

		DCD	_LAST-MAPOFFSET
_TFORTH	DCB   6
		DCB	"'FORTH"
		ALIGN	
TFORTH
		_PUSH
		ADD	R5,R3,#148
		_NEXT

; Common functions

;   FORTH       ( -- )
;               Make FORTH the context vocabulary.

		DCD	_TFORTH-MAPOFFSET
_FORTH		DCB  5
		DCB	"FORTH"
		ALIGN	
FORTH
		_NEST
		BL      TFORTH
		BL	CNTXT
		BL	STORE
		_UNNEST

;   WITHIN      ( u ul uh -- t )
;               Return true if u is within the range of ul and uh.

		DCD	_FORTH-MAPOFFSET
_WITHI		DCB   6
		DCB	"WITHIN"
		ALIGN	
WITHI
		_NEST
		BL	OVER
		BL	SUBB
		BL	TOR
		BL	SUBB
		BL	RFROM
		BL	ULESS
		_UNNEST

; Divide

;   UM/MOD      ( udl udh u -- ur uq )
;               Unsigned divide of a double by a single. Return mod and quotient.

		DCD	_WITHI-MAPOFFSET
_UMMOD		DCB   6
		DCB	"UM/MOD"
		ALIGN	
UMMOD
		MOV	R7,#1
		LDR	R4,[R1],#4
		LDR	R6,[R1]
UMMOD0		ADDS	R6,R6,R6
		ADCS	R4,R4,R4
		BCC	UMMOD1
		SUB	R4,R4,R5
		ADD	R6,R6,#1
		B UMMOD2
UMMOD1		SUBS	R4,R4,R5
		ADDCS	R6,R6,#1
		BCS	UMMOD2
		ADD	R4,R4,R5
UMMOD2		ADDS	R7,R7,R7
		BCC	UMMOD0
		MOV	R5,R6
		STR	R4,[R1]
		_NEXT

;   M/MOD       ( d n -- r q )
;               Signed floored divide of double by single. Return mod and quotient.

		DCD	_UMMOD-MAPOFFSET
_MSMOD		DCB  5
		DCB	"M/MOD"
		ALIGN	
MSMOD	
	_NEST
		BL      DUPP
		BL	ZLESS
		BL	DUPP
		BL	TOR
		BL      QBRAN
		DCD	MMOD1-MAPOFFSET
		BL      NEGAT
		BL	TOR
		BL	DNEGA
		BL	RFROM
MMOD1          BL      TOR
		BL	DUPP
		BL	ZLESS
		BL      QBRAN
		DCD	MMOD2-MAPOFFSET
		BL      RAT
		BL	PLUS
MMOD2          BL      RFROM
		BL	UMMOD
		BL	RFROM
		BL      QBRAN
		DCD	MMOD3-MAPOFFSET
		BL      SWAP
		BL	NEGAT
		BL	SWAP
MMOD3   
       _UNNEST

;   /MOD        ( n n -- r q )
;               Signed divide. Return mod and quotient.

		DCD	_MSMOD-MAPOFFSET
_SLMOD		DCB   4
		DCB	"/MOD"
		ALIGN	
SLMOD
		_NEST
		BL	OVER
		BL	ZLESS
		BL	SWAP
		BL	MSMOD
		_UNNEST

;   MOD         ( n n -- r )
;               Signed divide. Return mod only.

		DCD	_SLMOD-MAPOFFSET
_MODD		DCB  3
		DCB	"MOD"
		ALIGN	
MODD
		_NEST
		BL      SLMOD
		BL	DROP
		_UNNEST

;   /           ( n n -- q )
;               Signed divide. Return quotient only.

		DCD	_MODD-MAPOFFSET
_SLASH		DCB  1
		DCB	"/"
		ALIGN	
SLASH
		_NEST
		BL      SLMOD
		BL	SWAP
		BL	DROP
		_UNNEST

;   */MOD       ( n1 n2 n3 -- r q )
;               Multiply n1 and n2, then divide by n3. Return mod and quotient.

		DCD	_SLASH-MAPOFFSET
_SSMOD		DCB  5
		DCB	"*/MOD"
		ALIGN	
SSMOD
		_NEST
		BL      TOR
		BL	MSTAR
		BL	RFROM
		BL	MSMOD
		_UNNEST

;   */          ( n1 n2 n3 -- q )
;               Multiply n1 by n2, then divide by n3. Return quotient only.

		DCD	_SSMOD-MAPOFFSET
_STASL		DCB  2
		DCB	"*/"
		ALIGN	
STASL
		_NEST
		BL      SSMOD
		BL	SWAP
		BL	DROP
		_UNNEST

; Miscellaneous

;   ALIGNED     ( b -- a )
;               Align address to the cell boundary.

		DCD	_STASL-MAPOFFSET
_ALGND		DCB   7
		DCB	"ALIGNED"
		ALIGN	
ALGND
		ADD     R5,R5,#3
		MVN     R4,#3
		AND	R5,R5,R4
		_NEXT

;   >CHAR       ( c -- c )
;               Filter non-printing characters.

		DCD	_ALGND-MAPOFFSET
_TCHAR		DCB  5
		DCB	">CHAR"
		ALIGN	
TCHAR
		_NEST
		_DOLIT
		DCD	0x7F
		BL	ANDD
		BL	DUPP		;mask msb
		BL      BLANK
		_DOLIT
		DCD	127
		BL	WITHI		;check for printable
		BL      INVER
		BL	QBRAN
		DCD	TCHA1
		BL      DROP
		_DOLIT
		DCD	'_'		;replace non-printables
TCHA1
          _UNNEST

;   DEPTH       ( -- n )
;               Return the depth of the data stack.

		DCD	_TCHAR-MAPOFFSET
_DEPTH		DCB  5
		DCB	"DEPTH"
		ALIGN	
DEPTH
		_NEST
		BL      SPAT
		BL	SZERO
		BL	AT
		BL	SWAP
		BL	SUBB
		BL      CELLSL
		BL	ONEM
		_UNNEST

;   PICK        ( ... +n -- ... w )
;               Copy the nth stack item to tos.

		DCD	_DEPTH-MAPOFFSET
_PICK		DCB  4
		DCB	"PICK"
		ALIGN	
PICK
		_NEST
		BL	ONEP
		BL	CELLS
		BL	SPAT
		BL	PLUS
		BL	AT
		_UNNEST

; Memory access

;   HERE        ( -- a )
;               Return the top of the code dictionary.

		DCD	_PICK-MAPOFFSET
_HERE		DCB  4
		DCB	"HERE"
		ALIGN	
HERE
		_NEST
		BL      CP
		BL	AT
		_UNNEST

;   PAD         ( -- a )
;               Return the address of a temporary buffer.

		DCD	_HERE-MAPOFFSET
_PAD		DCB  3
		DCB	"PAD"
		ALIGN	
PAD
		_NEST
		BL      HERE
		ADD	R5,R5,#80
		_UNNEST

;   TIB         ( -- a )
;               Return the address of the terminal input buffer.

		DCD	_PAD-MAPOFFSET
_TIB		DCB  3
		DCB	"TIB"
		ALIGN	
TIB
		_NEST
		BL      NTIB
		BL	CELLP
		BL	AT
		_UNNEST

;   @EXECUTE    ( a -- )
;               Execute vector stored in address a.

		DCD	_TIB-MAPOFFSET
_ATEXE		DCB   8
		DCB	"@EXECUTE"
		ALIGN	
ATEXE
		MOVS	R4,R5
		_POP
		LDR	R4,[R4]
		BXNE	R4
		_NEXT

;   CMOVE       ( b1 b2 u -- )
;               Copy u bytes from b1 to b2.

		DCD	_ATEXE-MAPOFFSET
_CMOVE		DCB   5
		DCB	"CMOVE"
		ALIGN	
CMOVE
		LDR	R6,[R1],#4
		LDR	R7,[R1],#4
		B CMOV1
CMOV0		LDRB	R4,[R7],#1
		STRB	R4,[R6],#1
CMOV1		MOVS	R5,R5
		BEQ	CMOV2
		SUB	R5,R5,#1
		B CMOV0
CMOV2
		_POP
		_NEXT

;   MOVE       ( a1 a2 u -- )
;               Copy u bytes from a1 to a2.

		DCD	_CMOVE-MAPOFFSET
_MOVE		DCB   4
		DCB	"MOVE"
		ALIGN	
MOVE	AND	R5,R5,#-4
		LDR	R6,[R1],#4
		LDR	R7,[R1],#4
		B MOVE1
MOVE0		LDR	R4,[R7],#4
		STR	R4,[R6],#4
MOVE1		MOVS	R5,R5
		BEQ	MOVE2
		SUB	R5,R5,#4
		B MOVE0
MOVE2
		_POP
		_NEXT

;   FILL        ( b u c -- )
;               Fill u bytes of character c to area beginning at b.

		DCD	_MOVE-MAPOFFSET
_FILL		DCB   4
		DCB	"FILL"
		ALIGN	
FILL
		LDR	R6,[R1],#4
		LDR	R7,[R1],#4
FILL0		B FILL1
		MOV	R5,R5
FILL1		STRB	R5,[R7],#1
		MOVS	R6,R6
		BEQ	FILL2
		SUB	R6,R6,#1
		B FILL0
FILL2
		_POP
		_NEXT

;   -TRAILING   ( b u -- b u )
;               Adjust the count to eliminate trailing white space.

		DCD	_FILL-MAPOFFSET
_DTRAI		DCB  9
		DCB	"-TRAILING"
		ALIGN	
DTRAI		_NEST
		BL      TOR
		B	DTRA2
DTRA1          BL      BLANK
		BL	OVER
		BL	RAT
		BL	PLUS
		BL	CAT
		BL	LESS
		BL      QBRAN
		DCD	DTRA2-MAPOFFSET
		BL      RFROM
		BL	ONEP
		_UNNEST ;adjusted count
          BL      DONXT
		DCD	DTRA1-MAPOFFSET
DTRA2
		_DOLIT
		DCD	0
		_UNNEST			;count=0

;   PACK_       ( b u a -- a )
;               Build a countedDCB with u characters from b. Null fill.

		DCD	_DTRAI-MAPOFFSET
_PACKS		DCB  5
		DCB	"PACK$$"
		ALIGN	
PACKS
		_NEST
		BL      ALGND
		BL	DUPP
		BL	TOR		;strings only on cell boundary
		BL      OVER
		BL	PLUS
		BL	ONEP
		_DOLIT
		DCD	0xFFFFFFFC
		BL	ANDD		;count mod cell
		_DOLIT
		DCD	0
		BL	SWAP
		BL	STORE		;null fill cell
		BL      RAT
		BL	DDUP
		BL	CSTOR
		BL	ONEP		;save count
		BL      SWAP
		BL	CMOVE
		BL	RFROM
		_UNNEST   ;moveDCB

; Numeric output, single precision

;   DIGIT       ( u -- c )
;               Convert digit u to a character.

		DCD	_PACKS-MAPOFFSET
_DIGIT		DCB  5
		DCB	"DIGIT"
		ALIGN	
DIGIT
		_NEST
		_DOLIT
		DCD	9
		BL	OVER
		BL	LESS
		AND	R5,R5,#7
		BL	PLUS
		ADD	R5,R5,#'0'
		_UNNEST

;   EXTRACT     ( n base -- n c )
;               Extract the least significant digit from n.

		DCD	_DIGIT-MAPOFFSET
_EXTRC		DCB  7
		DCB	"EXTRACT"
		ALIGN	
EXTRC
		_NEST
		_DOLIT
		DCD	0
		BL	SWAP
		BL	UMMOD
		BL      SWAP
		BL	DIGIT
		_UNNEST

;   <#          ( -- )
;               Initiate the numeric output process.

		DCD	_EXTRC-MAPOFFSET
_BDIGS		DCB  2
		DCB	"<#"
		ALIGN	
BDIGS
		_NEST
		BL      PAD
		BL	HLD
		BL	STORE
		_UNNEST

;   HOLD        ( c -- )
;               Insert a character into the numeric outputDCB.

		DCD	_BDIGS-MAPOFFSET
_HOLD		DCB  4
		DCB	"HOLD"
		ALIGN	
HOLD
		_NEST
		BL      HLD
		BL	AT
		BL	ONEM
		BL      DUPP
		BL	HLD
		BL	STORE
		BL	CSTOR
		_UNNEST

;   #           ( u -- u )
;               Extract one digit from u and append the digit to outputDCB.

		DCD	_HOLD-MAPOFFSET
_DIG		DCB  1
		DCB	"#"
		ALIGN	
DIG
		_NEST
		BL      BASE
		BL	AT
		BL	EXTRC
		BL	HOLD
		_UNNEST

;   #S          ( u -- 0 )
;               Convert u until all digits are added to the outputDCB.

		DCD	_DIG-MAPOFFSET
_DIGS		DCB  2
		DCB	"#S"
		ALIGN	
DIGS
		_NEST
DIGS1          BL      DIG
		BL	DUPP
		BL      QBRAN
		DCD	DIGS2-MAPOFFSET
		B	DIGS1
DIGS2
          _UNNEST

;   SIGN        ( n -- )
;               Add a minus sign to the numeric outputDCB.

		DCD	_DIGS-MAPOFFSET
_SIGN		DCB  4
		DCB	"SIGN"
		ALIGN	
SIGN
		_NEST
		BL      ZLESS
		BL      QBRAN
		DCD	SIGN1-MAPOFFSET
		_DOLIT
		DCD	'-'
		BL	HOLD
SIGN1
          _UNNEST

;   #>          ( w -- b u )
;               Prepare the outputDCB to be TYPE'd.

		DCD	_SIGN-MAPOFFSET
_EDIGS		DCB  2
		DCB	"#>"
		ALIGN	
EDIGS
		_NEST
		BL      DROP
		BL	HLD
		BL	AT
		BL      PAD
		BL	OVER
		BL	SUBB
		_UNNEST

;   str         ( n -- b u )
;               Convert a signed integer to a numericDCB.

		DCD	_EDIGS-MAPOFFSET
_STRR		DCB  3
		DCB	"str"
		ALIGN	
STRR
		_NEST
		BL      DUPP
		BL	TOR
		BL	ABSS
		BL      BDIGS
		BL	DIGS
		BL	RFROM
		BL      SIGN
		BL	EDIGS
		_UNNEST

;   HEX         ( -- )
;               Use radix 16 as base for numeric conversions.

		DCD	_STRR-MAPOFFSET
_HEX		DCB  3
		DCB	"HEX"
		ALIGN	
HEX
		_NEST
		_DOLIT
		DCD	16
		BL	BASE
		BL	STORE
		_UNNEST

;   DECIMAL     ( -- )
;               Use radix 10 as base for numeric conversions.

		DCD	_HEX-MAPOFFSET
_DECIM		DCB  7
		DCB	"DECIMAL"
		ALIGN	
DECIM
		_NEST
		_DOLIT
		DCD	10
		BL	BASE
		BL	STORE
		_UNNEST

; Numeric input, single precision

;   DIGIT?      ( c base -- u t )
;               Convert a character to its numeric value. A flag indicates success.

		DCD	_DECIM-MAPOFFSET
_DIGTQ		DCB  6
		DCB	"DIGIT?"
		ALIGN	
DIGTQ
		_NEST
		BL      TOR
		_DOLIT
		DCD	'0'
		BL	SUBB
		_DOLIT
		DCD	9
		BL	OVER
		BL	LESS
		BL      QBRAN
		DCD	DGTQ1
		_DOLIT
		DCD	7
		BL	SUBB
		BL      DUPP
		_DOLIT
		DCD	10
		BL	LESS
		BL	ORR
DGTQ1          BL      DUPP
		BL	RFROM
		BL	ULESS
		_UNNEST

;   NUMBER?     ( a -- n T | a F )
;               Convert a numberDCB to integer. Push a flag on tos.

		DCD	_DIGTQ-MAPOFFSET
_NUMBQ		DCB  7
		DCB	"NUMBER?"
		ALIGN	
NUMBQ
		_NEST
		BL      BASE
		BL	AT
		BL	TOR
		_DOLIT
		DCD	0
		BL	OVER
		BL	COUNT
		BL      OVER
		BL	CAT
		_DOLIT
		DCD	'_'
		BL	EQUAL
		BL      QBRAN
		DCD	NUMQ1-MAPOFFSET
		BL      HEX
		BL	SWAP
		BL	ONEP
		BL      SWAP
		BL	ONEM
NUMQ1          BL      OVER
		BL	CAT
		_DOLIT
		DCD	'-'
		BL	EQUAL
		BL	TOR
		BL      SWAP
		BL	RAT
		BL	SUBB
		BL	SWAP
		BL	RAT
		BL	PLUS
		BL	QDUP
		BL      QBRAN
		DCD	NUMQ6-MAPOFFSET
		BL      ONEM
		BL	TOR
NUMQ2          BL      DUPP
		BL	TOR
		BL	CAT
		BL	BASE
		BL	AT
		BL	DIGTQ
		BL      QBRAN
		DCD	NUMQ4-MAPOFFSET
		BL      SWAP
		BL	BASE
		BL	AT
		BL	STAR
		BL	PLUS
		BL	RFROM
		BL      ONEP
		BL      DONXT
		DCD	NUMQ2-MAPOFFSET
		BL      RAT
		BL	SWAP
		BL	DROP
		BL      QBRAN
		DCD	NUMQ3-MAPOFFSET
		BL      NEGAT
NUMQ3          BL      SWAP
		B      NUMQ5
NUMQ4          BL      RFROM
		BL	RFROM
		BL	DDROP
		BL	DDROP
		_DOLIT
		DCD	0
NUMQ5          BL      DUPP
NUMQ6          BL      RFROM
		BL	DDROP
		BL      RFROM
		BL	BASE
		BL	STORE
		_UNNEST

; Basic I/O

;   ?KEY        ( -- c T | F )
;               Return input character and true, or a false if no input.

		DCD	_NUMBQ-MAPOFFSET
_QKEY		DCB  4
		DCB	"?KEY"
		ALIGN	
QKEY
		_NEST
		BL      TQKEY
		BL	ATEXE
		_UNNEST

;   KEY         ( -- c )
;               Wait for and return an input character.

		DCD	_QKEY-MAPOFFSET
_KEY		DCB  3
		DCB	"KEY"
		ALIGN	
KEY
		_NEST
KEY1           BL      QKEY
		BL      QBRAN
		DCD	KEY1-MAPOFFSET
		_UNNEST

;   EMIT        ( c -- )
;               Send a character to the output device.

		DCD	_KEY-MAPOFFSET
_EMIT		DCB  4
		DCB	"EMIT"
		ALIGN	
EMIT
		_NEST
		BL	TEMIT
		BL	ATEXE
		_UNNEST

;   NUF?        ( -- t )
;               Return false if no input, else pause and if CR return true.

		DCD	_EMIT-MAPOFFSET
_NUFQ		DCB  4
		DCB	"NUF?"
		ALIGN	
NUFQ
		_NEST
		BL      QKEY
		BL	DUPP
		BL      QBRAN
		DCD	NUFQ1-MAPOFFSET
		BL      DDROP
		BL	KEY
		_DOLIT
		DCD	ERR
		BL	EQUAL
NUFQ1
          _UNNEST

;   PACE        ( -- )
;               Send a pace character for the file downloading process.

		DCD	_NUFQ-MAPOFFSET
_PACE		DCB  4
		DCB	"PACE"
		ALIGN	
PACE
		_NEST
		_DOLIT
		DCD	11
		BL	EMIT
		_UNNEST

;   SPACE       ( -- )
;               Send the blank character to the output device.

		DCD	_PACE-MAPOFFSET
_SPACE		DCB  5
		DCB	"SPACE"
		ALIGN	
SPACE
		_NEST
		BL      BLANK
		BL	EMIT
		_UNNEST

;   SPACES      ( +n -- )
;               Send n spaces to the output device.

		DCD	_SPACE-MAPOFFSET
_SPACS		DCB  6
		DCB	"SPACES"
		ALIGN	
SPACS
		_NEST
		_DOLIT
		DCD	0
		BL	MAX
		BL	TOR
		B	CHAR2
CHAR1  BL      SPACE
CHAR2  BL      DONXT
		DCD	CHAR1-MAPOFFSET
		_UNNEST

;   TYPE        ( b u -- )
;               Output u characters from b.

		DCD	_SPACS-MAPOFFSET
_TYPEE		DCB	4
		DCB	"TYPE"
		ALIGN	
TYPEE
		_NEST
		BL  TOR
		B	TYPE2
TYPE1  BL  COUNT
		BL	EMIT
TYPE2  BL  DONXT
		DCD	TYPE1-MAPOFFSET
		BL      DROP
		_UNNEST

;   CR          ( -- )
;               Output a carriage return and a line feed.

		DCD	_TYPEE-MAPOFFSET
_CR		DCB  2
		DCB	"CR"
		ALIGN	
CR
		_NEST
		_DOLIT
		DCD	CRR
		BL	EMIT
		_DOLIT
		DCD	LF
		BL	EMIT
		_UNNEST

;   do_$        ( -- a )
;               Return the address of a compiledDCB.

		DCD	_CR-MAPOFFSET
_DOSTR		DCB  COMPO+3
		DCB	"do$$"
		ALIGN	
DOSTR
		_NEST
		BL      RFROM
		BL	RAT
		BL	RFROM
		BL	COUNT
		BL	PLUS
		BL      ALGND
		BL	TOR
		BL	SWAP
		BL	TOR
		_UNNEST

;   _"|		( -- a )
;               Run time routine compiled by _". Return address of a compiledDCB.

		DCD	_DOSTR-MAPOFFSET
_STRQP		DCB  COMPO+3
		DCB	"$$""|"
		ALIGN	
STRQP
		_NEST
		BL      DOSTR
		_UNNEST			;force a call to do_

;   ._		( -- )
;               Run time routine of ." . Output a compiledDCB.

		DCD	_STRQP-MAPOFFSET
_DOTST		DCB  COMPO+2
		DCB	".$$"
		ALIGN	
DOTST
		_NEST
		BL	COUNT
		BL	TYPEE
		_UNNEST

;   ."|		( -- )
;               Run time routine of ." . Output a compiledDCB.

		DCD	_DOTST-MAPOFFSET
_DOTQP		DCB  COMPO+3
		DCB	".""|"
		ALIGN	
DOTQP
		_NEST
		BL      DOSTR
		BL	DOTST
		_UNNEST

;   .R          ( n +n -- )
;               Display an integer in a field of n columns, right justified.

		DCD	_DOTQP-MAPOFFSET
_DOTR		DCB  2
		DCB	".R"
		ALIGN	
DOTR
		_NEST
		BL      TOR
		BL	STRR
		BL	RFROM
		BL	OVER
		BL	SUBB
		BL      SPACS
		BL	TYPEE
		_UNNEST

;   U.R         ( u +n -- )
;               Display an unsigned integer in n column, right justified.

		DCD	_DOTR-MAPOFFSET
_UDOTR		DCB  3
		DCB	"U.R"
		ALIGN	
UDOTR
		_NEST
		BL      TOR
		BL	BDIGS
		BL	DIGS
		BL	EDIGS
		BL      RFROM
		BL	OVER
		BL	SUBB
		BL      SPACS
		BL	TYPEE
		_UNNEST

;   U.          ( u -- )
;               Display an unsigned integer in free format.

		DCD	_UDOTR-MAPOFFSET
_UDOT		DCB  2
		DCB	"U."
		ALIGN	
UDOT
		_NEST
		BL      BDIGS
		BL	DIGS
		BL	EDIGS
		BL      SPACE
		BL	TYPEE
		_UNNEST

;   .           ( w -- )
;               Display an integer in free format, preceeded by a space.

		DCD	_UDOT-MAPOFFSET
_DOT		DCB  1
		DCB	"."
		ALIGN	
DOT
		_NEST
		BL      BASE
		BL	AT
		_DOLIT
		DCD	10
		BL	XORR		;?decimal
		BL      QBRAN
		DCD	DOT1
		BL      UDOT
		_UNNEST			;no,display unsigned
DOT1           BL      STRR
		BL	SPACE
		BL	TYPEE
		_UNNEST			;yes, display signed

;   ?           ( a -- )
;               Display the contents in a memory cell.

		DCD	_DOT-MAPOFFSET
_QUEST		DCB  1
		DCB	"?"
		ALIGN	
QUEST
		_NEST
		BL      AT
		BL	DOT
		_UNNEST

; Parsing

;   parse       ( b u c -- b u delta ; DCB> )
;               ScanDCB delimited by c. Return foundDCB and its offset.

		DCD	_QUEST-MAPOFFSET
_PARS		DCB  5
		DCB	"parse"
		ALIGN	
PARS
		_NEST
		BL      TEMP
		BL	STORE
		BL	OVER
		BL	TOR
		BL	DUPP
		BL      QBRAN
		DCD	PARS8-MAPOFFSET
		BL      ONEM
		BL	TEMP
		BL	AT
		BL	BLANK
		BL	EQUAL
		BL      QBRAN
		DCD	PARS3-MAPOFFSET
		BL      TOR
PARS1          BL      BLANK
		BL	OVER
		BL	CAT		;skip leading blanks ONLY
		BL      SUBB
		BL	ZLESS
		BL	INVER
		BL      QBRAN
		DCD	PARS2-MAPOFFSET
		BL      ONEP
		BL      DONXT
		DCD	PARS1-MAPOFFSET
		BL      RFROM
		BL	DROP
		_DOLIT
		DCD	0
		BL	DUPP
		_UNNEST
PARS2          BL      RFROM
PARS3          BL      OVER
		BL	SWAP
		BL      TOR
PARS4          BL      TEMP
		BL	AT
		BL	OVER
		BL	CAT
		BL	SUBB		;scan for delimiter
		BL      TEMP
		BL	AT
		BL	BLANK
		BL	EQUAL
		BL      QBRAN
		DCD	PARS5-MAPOFFSET
		BL      ZLESS
PARS5          BL      QBRAN
		DCD	PARS6-MAPOFFSET
		BL      ONEP
		BL      DONXT
		DCD	PARS4-MAPOFFSET
		BL      DUPP
		BL	TOR
		B      PARS7
PARS6          BL      RFROM
		BL	DROP
		BL	DUPP
		BL      ONEP
		BL	TOR
PARS7          BL      OVER
		BL	SUBB
		BL      RFROM
		BL	RFROM
		BL	SUBB
		_UNNEST
PARS8          BL      OVER
		BL	RFROM
		BL	SUBB
		_UNNEST

;   PARSE       ( c -- b u ; DCB> )
;               Scan input stream and return countedDCB delimited by c.

		DCD	_PARS-MAPOFFSET
_PARSE		DCB  5
		DCB	"PARSE"
		ALIGN	
PARSE
		_NEST
		BL      TOR
		BL	TIB
		BL	INN
		BL	AT
		BL	PLUS		;current input buffer pointer
		BL      NTIB
		BL	AT
		BL	INN
		BL	AT
		BL	SUBB		;remaining count
		BL      RFROM
		BL	PARS
		BL	INN
		BL	PSTOR
		_UNNEST

;   .(          ( -- )
;               Output followingDCB up to next ) .

		DCD	_PARSE-MAPOFFSET
_DOTPR		DCB  IMEDD+2
		DCB	".("
		ALIGN	
DOTPR
		_NEST
		_DOLIT
		DCD	')'
		BL	PARSE
		BL	TYPEE
		_UNNEST

;   (           ( -- )
;               Ignore followingDCB up to next ) . A comment.

		DCD	_DOTPR-MAPOFFSET
_PAREN		DCB  IMEDD+1
		DCB	"("
		ALIGN	
PAREN		_NEST
		_DOLIT
		DCD	')'
		BL	PARSE
		BL	DDROP
		_UNNEST

;   \           ( -- )
;               Ignore following text till the end of line.

		DCD	_PAREN-MAPOFFSET
_BKSLA		DCB  IMEDD+1
		DCB	"\\"
		ALIGN	
BKSLA
		_NEST
		BL      NTIB
		BL	AT
		BL	INN
		BL	STORE
		_UNNEST

;   CHAR        ( -- c )
;               Parse next word and return its first character.

		DCD	_BKSLA-MAPOFFSET
_CHAR		DCB  4
		DCB	"CHAR"
		ALIGN	
CHAR
		_NEST
		BL      BLANK
		BL	PARSE
		BL	DROP
		BL	CAT
		_UNNEST

;   WORD        ( c -- a ; DCB> )
;               Parse a word from input stream and copy it to code dictionary.

		DCD	_CHAR-MAPOFFSET
_WORDD		DCB  4
		DCB	"WORD"
		ALIGN	
WORDD
		_NEST
		BL      PARSE
		BL	HERE
		BL	CELLP
		BL	PACKS
		_UNNEST

;   TOKEN       ( -- a ; DCB> )
;               Parse a word from input stream and copy it to name dictionary.

		DCD	_WORDD-MAPOFFSET
_TOKEN		DCB  5
		DCB	"TOKEN"
		ALIGN	
TOKEN
		_NEST
		BL      BLANK
		BL	WORDD
		_UNNEST

; Dictionary search

;   NAME>       ( na -- ca )
;               Return a code address given a name address.

		DCD	_TOKEN-MAPOFFSET
_NAMET		DCB  5
		DCB	"NAME>"
		ALIGN	
NAMET
		_NEST
		BL      COUNT
		_DOLIT
		DCD	0x1F
		BL	ANDD
		BL	PLUS
		BL	ALGND
		_UNNEST

;   SAME?       ( a a u -- a a f \ -0+ )
;               Compare u cells in twoDCBs. Return 0 if identical.

		DCD	_NAMET-MAPOFFSET
_SAMEQ		DCB  5
		DCB	"SAME?"
		ALIGN	
SAMEQ
		_NEST
		BL      TOR
		B      SAME2
SAME1          BL      OVER
		BL	RAT
		BL	CELLS
		BL	PLUS
		BL	AT		;32/16 mix-up
		BL      OVER
		BL	RAT
		BL	CELLS
		BL	PLUS
		BL	AT		;32/16 mix-up
		BL      SUBB
		BL	QDUP
		BL      QBRAN
		DCD	SAME2-MAPOFFSET
		BL      RFROM
		BL	DROP
		_UNNEST			;strings not equal
SAME2          BL      DONXT
		DCD	SAME1-MAPOFFSET
		_DOLIT
		DCD	0
		_UNNEST			;strings equal

;   find        ( a va -- ca na | a F )
;               Search a vocabulary for aDCB. Return ca and na if succeeded.

		DCD	_SAMEQ-MAPOFFSET
_FIND		DCB  4
		DCB	"find"
		ALIGN	
FIND
		_NEST
		BL      SWAP
		BL	DUPP
		BL	CAT
		BL      CELLSL
		BL	TEMP
		BL	STORE		;32/16 bit mix-up
		BL      DUPP
		BL	AT
		BL	TOR
		BL	CELLP
		BL	SWAP
FIND1          BL      AT
		BL	DUPP
		BL      QBRAN
		DCD	FIND6-MAPOFFSET
		BL      DUPP
		BL	AT
		_DOLIT
		DCD	MASKK
		BL	ANDD
		BL	RAT
		BL	XORR
		BL      QBRAN
		DCD	FIND2-MAPOFFSET
		BL      CELLM		;backup to link field
		B	FIND1		;try next word
FIND2          BL      CELLP
		BL	TEMP
		BL	AT
		BL	SAMEQ		;32/16 mix-up 
FIND3          B	FIND4
FIND6          BL      RFROM
		BL	DROP
		BL      SWAP
		BL	CELLM
		BL	SWAP
		_UNNEST			;not in this vocab
FIND4          BL      QBRAN
		DCD	FIND5-MAPOFFSET
		BL      CELLM
		BL	CELLM
		B	FIND1
FIND5          BL      RFROM
		BL	DROP
		BL	SWAP
		BL	DROP
		BL      CELLM		;na
		BL      DUPP
		BL	NAMET
		BL	SWAP
		_UNNEST			;ca
		DCD	0,0,0,0,0,0,0,0,0	;patch to match ADuCef.hex

;   NAME?       ( a -- ca na | a F )
;               Search all context vocabularies for aDCB.

		DCD	_FIND-MAPOFFSET
_NAMEQ		DCB  5
		DCB	"NAME?"
		ALIGN	
NAMEQ
		_NEST
		BL      CNTXT
		BL	DUPP
		BL	DAT
		BL	XORR		;?context=also
		BL      QBRAN
		DCD	NAMQ1-MAPOFFSET
		BL      CELLM		;no, start with context
NAMQ1          BL      TOR
NAMQ2          BL      RFROM
		BL	CELLP
		BL	DUPP
		BL	TOR		;next in search order
		BL      AT
		BL	QDUP
		BL      QBRAN
		DCD	NAMQ3-MAPOFFSET
		BL      FIND
		BL	QDUP		;search vocabulary
		BL      QBRAN
		DCD	NAMQ2-MAPOFFSET
		BL      RFROM
		BL	DROP
		_UNNEST			;found name
NAMQ3          BL      RFROM
		BL	DROP		;name not found
		_DOLIT
		DCD	0
		_UNNEST			;false flag

; Terminal response

;   ^H          ( bot eot cur -- bot eot cur )
;               Backup the cursor by one character.

		DCD	_NAMEQ-MAPOFFSET
_BKSP		DCB  2
		DCB	"^H"
		ALIGN	
BKSP
		_NEST
		BL      TOR
		BL	OVER
		BL	RFROM
		BL	SWAP
		BL	OVER
		BL	XORR
		BL      QBRAN
		DCD	BACK1-MAPOFFSET
		_DOLIT
		DCD	BKSPP
		BL	TECHO
		BL	ATEXE
		BL	ONEM
		BL      BLANK
		BL	TECHO
		BL	ATEXE
		_DOLIT
		DCD	BKSPP
		BL	TECHO
		BL	ATEXE
BACK1
          _UNNEST

;   TAP         ( bot eot cur c -- bot eot cur )
;               Accept and echo the key stroke and bump the cursor.

		DCD	_BKSP-MAPOFFSET
_TAP		DCB  3
		DCB	"TAP"
		ALIGN	
TAP
		_NEST
		BL      DUPP
		BL	TECHO
		BL	ATEXE
		BL      OVER
		BL	CSTOR
		BL	ONEP
		_UNNEST

;   kTAP        ( bot eot cur c -- bot eot cur )
;               Process a key stroke, CR or backspace.

		DCD	_TAP-MAPOFFSET
_KTAP		DCB  4
		DCB	"kTAP"
		ALIGN	
KTAP
		_NEST
		BL      DUPP
		_DOLIT
		DCD	CRR
		BL	XORR
		BL      QBRAN
		DCD	KTAP2-MAPOFFSET
		_DOLIT
		DCD	BKSPP
		BL	XORR
		BL      QBRAN
		DCD	KTAP1-MAPOFFSET
		BL      BLANK
		BL	TAP
		_UNNEST
		DCD	0	;patch
KTAP1          BL      BKSP
		_UNNEST
KTAP2          BL      DROP
		BL	SWAP
		BL	DROP
		BL	DUPP
		_UNNEST

;   accept      ( b u -- b u )
;               Accept characters to input buffer. Return with actual count.

		DCD	_KTAP-MAPOFFSET
_ACCEP		DCB  6
		DCB	"accept"
		ALIGN	
ACCEP
		_NEST
		BL      OVER
		BL	PLUS
		BL	OVER
ACCP1          BL      DDUP
		BL	XORR
		BL      QBRAN
		DCD	ACCP4-MAPOFFSET
		BL      KEY
		BL	DUPP
		BL      BLANK
		_DOLIT
		DCD	127
		BL	WITHI
		BL      QBRAN
		DCD	ACCP2-MAPOFFSET
		BL      TAP
		B      ACCP3
ACCP2          BL      TTAP
		BL	ATEXE
ACCP3          
		B      ACCP1
ACCP4          BL      DROP
		BL	OVER
		BL	SUBB
		_UNNEST

;   EXPECT      ( b u -- )
;               Accept input stream and store count in SPAN.

		DCD	_ACCEP-MAPOFFSET
_EXPEC		DCB  6
		DCB	"EXPECT"
		ALIGN	
EXPEC
		_NEST
		BL      TEXPE
		BL	ATEXE
		BL	SPAN
		BL	STORE
		BL	DROP
		_UNNEST

;   QUERY       ( -- )
;               Accept input stream to terminal input buffer.

		DCD	_EXPEC-MAPOFFSET
_QUERY		DCB  5
		DCB	"QUERY"
		ALIGN	
QUERY
		_NEST
		BL      TIB
		_DOLIT
		DCD	80
		BL	TEXPE
		BL	ATEXE
		BL	NTIB
		BL	STORE
		BL      DROP
		_DOLIT
		DCD	0
		BL	INN
		BL	STORE
		_UNNEST

; Error handling

;   CATCH       ( ca -- 0 | err# )
;               Execute word at ca and set up an error frame for it.

		DCD	_QUERY-MAPOFFSET
_CATCH		DCB  5
		DCB	"CATCH"
		ALIGN	
CATCH
		_NEST
		BL  SPAT
		BL	TOR
		BL	HANDL
		BL	AT
		BL	TOR		;save error frame
		BL  RPAT
		BL	HANDL
		BL	STORE
		BL	EXECU		;execute
		BL  RFROM
		BL	HANDL
		BL	STORE		;restore error frame
		BL  RFROM
		BL	DROP
		_DOLIT
		DCD	0
		_UNNEST ;no error

;   THROW       ( err# -- err# )
;               Reset system to current local error frame an update error flag.

		DCD	_CATCH-MAPOFFSET
_THROW		DCB  5
		DCB	"THROW"
		ALIGN	
THROW
		_NEST
		BL      HANDL
		BL	AT
		BL	RPSTO		;restore return stack
		BL      RFROM
		BL	HANDL
		BL	STORE		;restore handler frame
		BL      RFROM
		BL	SWAP
		BL	TOR
		BL	SPSTO		;restore data stack
		BL      DROP
		BL	RFROM
		_UNNEST

;   NULL_       ( -- a )
;               Return address of a nullDCB with zero count.

		DCD	_THROW-MAPOFFSET
_NULLS		DCB  5
		DCB	"NULL$$"
		ALIGN	
NULLS
		_NEST
		BL	DOVAR			;emulate CREATE
		DCD      0
;		DCB      99,111,121,111,116,101
;		ALIGN	2

;   ABORT       ( -- )
;               Reset data stack and jump to QUIT.

		DCD	_NULLS-MAPOFFSET
_ABORT		DCB  5
		DCB	"ABORT"
		ALIGN	
ABORT
		_NEST
		BL      NULLS
		BL	THROW
		_UNNEST		

;   _abort"	( f -- )
;               Run time routine of ABORT" . Abort with a message.

		DCD	_ABORT-MAPOFFSET
_ABORQ		DCB  COMPO+6
		DCB	"abort\""
		ALIGN	
ABORQ
		_NEST
		BL      QBRAN
		DCD	ABOR1		;text flag
		BL      DOSTR
		BL	THROW		;pass errorDCB
ABOR1          BL      DOSTR
		BL	DROP
		_UNNEST			;drop error

; The text interpreter

;   _INTERPRET  ( a -- )
;               Interpret a word. If failed, try to convert it to an integer.

		DCD	_ABORQ-MAPOFFSET
_INTER		DCB  10
		DCB	"$$INTERPRET"
		ALIGN	
INTER
		_NEST
		BL      NAMEQ
		BL	QDUP		;?defined
		BL      QBRAN
		DCD	INTE1-MAPOFFSET
		BL      AT
		_DOLIT
		DCD	COMPO
		BL	ANDD		;?compile only lexicon bits
		BL      ABORQ
		DCB	13
		DCB	" compile only"
		ALIGN	
		BL      EXECU
		_UNNEST			;execute defined word
INTE1          BL      TNUMB
		BL	ATEXE		;convert a number
		BL      QBRAN
		DCD	INTE2-MAPOFFSET
		_UNNEST
INTE2          BL      THROW		;error
		_UNNEST

;   [           ( -- )
;               Start the text interpreter.

		DCD	_INTER-MAPOFFSET
_LBRAC		DCB  IMEDD+1
		DCB	"["
		ALIGN	
LBRAC
		_NEST
		_DOLIT
		DCD	INTER-MAPOFFSET
		BL	TEVAL
		BL	STORE
		_UNNEST

;   .OK         ( -- )
;               Display "ok" only while interpreting.

		DCD	_LBRAC-MAPOFFSET
_DOTOK		DCB  3
		DCB	".OK"
		ALIGN	
DOTOK
		_NEST
		_DOLIT
		DCD	INTER-MAPOFFSET
		BL	TEVAL
		BL	AT
		BL	EQUAL
		BL      QBRAN
		DCD	DOTO1-MAPOFFSET
		BL      DOTQP
		DCB	3
		DCB	" ok"
		ALIGN	
DOTO1          BL      CR
		_UNNEST

;   ?STACK      ( -- )
;               Abort if the data stack underflows.

		DCD	_DOTOK-MAPOFFSET
_QSTAC		DCB  6
		DCB	"?STACK"
		ALIGN	
QSTAC
		_NEST
		BL      DEPTH
		BL	ZLESS		;check only for underflow
		BL      ABORQ
		DCB	10
		DCB	" underflow"
		ALIGN	
		_UNNEST

;   EVAL        ( -- )
;               Interpret the input stream.

		DCD	_QSTAC-MAPOFFSET
_EVAL		DCB  4
		DCB	"EVAL"
		ALIGN	
EVAL
		_NEST
EVAL1          BL      TOKEN
		BL	DUPP
		BL	CAT		;?input stream empty
		BL      QBRAN
		DCD	EVAL2-MAPOFFSET
		BL      TEVAL
		BL	ATEXE
		BL	QSTAC		;evaluate input, check stack
		B	EVAL1
EVAL2          BL      DROP
		BL	TPROM
		BL	ATEXE
		_UNNEST			;prompt

; Shell

;   PRESET      ( -- )
;               Reset data stack pointer and the terminal input buffer.

		DCD	_EVAL-MAPOFFSET
_PRESE		DCB  6
		DCB	"PRESET"
		ALIGN	
PRESE
		_NEST
		BL      SZERO
		BL	AT
		BL	SPSTO
		_DOLIT
		DCD	TIBB
		BL	NTIB
		BL	CELLP
		BL	STORE
		_UNNEST

;   xio         ( a a a -- )
;               Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

		DCD	_PRESE-MAPOFFSET
_XIO		DCB  COMPO+3
		DCB	"xio"
		ALIGN	
XIO
		_NEST
		_DOLIT
		DCD	ACCEP-MAPOFFSET
		BL	TEXPE
		BL	STORE
		BL      TTAP
		BL	STORE
		BL      TECHO
		BL	STORE
		BL      TPROM
		BL	STORE
		_UNNEST

;   FILE        ( -- )
;               Select I/O vectors for file download.

		DCD	_XIO-MAPOFFSET
_FILE		DCB  4
		DCB	"FILE"
		ALIGN	
FILE
		_NEST
		_DOLIT
		DCD	PACE-MAPOFFSET
		_DOLIT
		DCD	DROP-MAPOFFSET
		_DOLIT
		DCD	KTAP-MAPOFFSET
		BL	XIO
		_UNNEST

;   HAND        ( -- )
;               Select I/O vectors for terminal interface.

		DCD	_FILE-MAPOFFSET
_HAND		DCB  4
		DCB	"HAND"
		ALIGN	
HAND
		_NEST
		_DOLIT
		DCD	DOTOK-MAPOFFSET
		_DOLIT
		DCD	EMIT-MAPOFFSET
		_DOLIT
		DCD	KTAP-MAPOFFSET
		BL	XIO
		_UNNEST

;   I/O         ( -- a )
;               Array to store default I/O vectors.

		DCD	_HAND-MAPOFFSET
_ISLO		DCB  3
		DCB	"I/O"
		ALIGN	
ISLO
		_NEST
		BL	DOVAR			;emulate CREATE
		DCD	QRX
		DCD	TXSTO		;default I/O vectors
					;for 32 bit system 3/16/92

;   CONSOLE     ( -- )
;               Initiate terminal interface.

		DCD	_ISLO-MAPOFFSET
_CONSO		DCB  7
		DCB	"CONSOLE"
		ALIGN	
CONSO
		_NEST
		BL      ISLO
		BL	DAT
		BL	TQKEY
		BL	DSTOR		;restore default I/O device
		BL      HAND	;interactive mode
;		BL      FILE	;GBA mode
		_UNNEST			;keyboard input

;   QUIT        ( -- )
;               Reset return stack pointer and start text interpreter.

		DCD	_CONSO-MAPOFFSET
_QUIT		DCB  4
		DCB	"QUIT"
		ALIGN	
QUIT
		_NEST
		BL      RZERO
		BL	AT
		BL	RPSTO		;reset return stack pointer
		BL	CONSO
QUIT1          BL      LBRAC		;start interpretation
QUIT2          BL      QUERY		;get input
		_DOLIT
		DCD	EVAL-MAPOFFSET
		BL	CATCH
		BL	QDUP		;evaluate input
		BL      QBRAN
		DCD	QUIT2-MAPOFFSET		;continue till error
		BL      TPROM
		BL	AT
		BL	SWAP		;save input device
		BL      CONSO
		BL	NULLS
		BL	OVER
		BL	XORR		;?display error message
		BL      QBRAN
		DCD	QUIT3-MAPOFFSET
		BL      SPACE
		BL	COUNT
		BL	TYPEE		;error message
		BL      DOTQP
		DCB	3
		DCB	" ? "		;error prompt
		ALIGN	2
		BL      CR
QUIT3
          _DOLIT
		DCD	DOTOK-MAPOFFSET
		BL	XORR		;?file input
		BL      QBRAN
		DCD	QUIT4-MAPOFFSET
		_DOLIT
		DCD	ERR
		BL	EMIT		;file error, tell host
QUIT4          BL      PRESE		;some cleanup
		B	QUIT1

; The compiler

;   '           ( -- ca )
;               Search context vocabularies for the next word in input stream.

		DCD	_QUIT-MAPOFFSET
_TICK		DCB  1
		DCB	"'"
		ALIGN	
TICK
		_NEST
		BL      TOKEN
		BL	NAMEQ		;?defined
		BL      QBRAN
		DCD	TICK1-MAPOFFSET
		_UNNEST			;yes, push code address
TICK1          BL      THROW		;no, error

;   ALLOT       ( n -- )
;               Allocate n bytes to the code dictionary.

		DCD	_TICK-MAPOFFSET
_ALLOT		DCB  5
		DCB	"ALLOT"
		ALIGN	
ALLOT
		_NEST
		BL      CP
		BL	PSTOR
		_UNNEST			;adjust code pointer

;   ,           ( w -- )
;               Compile an integer into the code dictionary.

		DCD	_ALLOT-MAPOFFSET
_COMMA		DCB  1,","
		ALIGN	
COMMA
		_NEST
		BL      HERE
		BL	DUPP
		BL	CELLP		;cell boundary
		BL      CP
		BL	STORE
		BL	STORE
		_UNNEST			;adjust code pointer, compile

;   [COMPILE]   ( -- ; DCB> )
;               Compile the next immediate word into code dictionary.

		DCD	_COMMA-MAPOFFSET
_BCOMP		DCB  IMEDD+9
		DCB	"[COMPILE]"
		ALIGN	
BCOMP
		_NEST
		BL      TICK
		BL	COMMA
		_UNNEST

;   COMPILE     ( -- )
;               Compile the next address in colon list to code dictionary.

		DCD	_BCOMP-MAPOFFSET
_COMPI		DCB  COMPO+7
		DCB	"COMPILE"
		ALIGN	
COMPI
		_NEST
		BL      RFROM
		BL	DUPP
		BL	AT
		BL	CALLC		;compile BL instruction
		BL      CELLP
		BL	TOR
		_UNNEST			;adjust return address

;   LITERAL     ( w -- )
;               Compile tos to code dictionary as an integer literal.

		DCD	_COMPI-MAPOFFSET
_LITER		DCB  IMEDD+7
		DCB	"LITERAL"
		ALIGN	
LITER
		_NEST
		BL      COMPI
		DCD	DOLIT-MAPOFFSET
		BL	COMMA
		_UNNEST

;   _,"		( -- )
;               Compile a literalDCB up to next " .

		DCD	_LITER-MAPOFFSET
_STRCQ		DCB  3
		DCB	"$$,"""
		ALIGN	
STRCQ
		_NEST
		_DOLIT
		DCD	-4
		BL	CP
		BL	PSTOR
		_DOLIT
		DCD	'\"'
		BL	WORDD		;moveDCB to code dictionary
		BL      COUNT
		BL	PLUS
		BL	ALGND		;calculate aligned end ofDCB
		BL      CP
		BL	STORE
		_UNNEST 		;adjust the code pointer

;   RECURSE     ( -- )
;               Make the current word available for compilation.

		DCD	_STRCQ-MAPOFFSET
_RECUR		DCB  IMEDD+7
		DCB	"RECURSE"
		ALIGN	
RECUR
		_NEST
		BL      LAST
		BL	AT
		BL	NAMET
		BL	COMMA
		_UNNEST

; Structures

;   FOR         ( -- a )
;               Start a FOR-NEXT loop structure in a colon definition.

		DCD	_RECUR-MAPOFFSET
_FOR		DCB  IMEDD+3
		DCB	"FOR"
		ALIGN	
FOR
		_NEST
		BL      COMPI
		DCD	TOR-MAPOFFSET
		BL	HERE
		_UNNEST

;   BEGIN       ( -- a )
;               Start an infinite or indefinite loop structure.

		DCD	_FOR-MAPOFFSET
_BEGIN		DCB  IMEDD+5
		DCB	"BEGIN"
		ALIGN	
BEGIN
		_NEST
		BL      HERE
		_UNNEST

;   NEXT        ( a -- )
;               Terminate a FOR-NEXT loop structure.

		DCD	_BEGIN-MAPOFFSET
_NEXT		DCB  IMEDD+4
		DCB	"NEXT"
		ALIGN	
NEXT
		_NEST
		BL      COMPI
		DCD	DONXT-MAPOFFSET
		BL	COMMA
		_UNNEST

;   UNTIL       ( a -- )
;               Terminate a BEGIN-UNTIL indefinite loop structure.

		DCD	_NEXT-MAPOFFSET
_UNTIL		DCB  IMEDD+5
		DCB	"UNTIL"
		ALIGN	
UNTIL
		_NEST
		BL      COMPI
		DCD	QBRAN-MAPOFFSET
		BL	COMMA
		_UNNEST

;   AGAIN       ( a -- )
;               Terminate a BEGIN-AGAIN infinite loop structure.

		DCD	_UNTIL-MAPOFFSET
_AGAIN		DCB  IMEDD+5
		DCB	"AGAIN"
		ALIGN	
AGAIN
		_NEST
		BL      COMPI
		DCD	BRAN-MAPOFFSET
		BL	COMMA
		_UNNEST

;   IF          ( -- A )
;               Begin a conditional branch structure.

		DCD	_AGAIN-MAPOFFSET
_IFF		DCB  IMEDD+2
		DCB	"IF"
		ALIGN	
IFF
		_NEST
		BL      COMPI
		DCD	QBRAN-MAPOFFSET
		BL	HERE
		_DOLIT
		DCD	0
		BL	COMMA
		_UNNEST

;   AHEAD       ( -- A )
;               Compile a forward branch instruction.

		DCD	_IFF-MAPOFFSET
_AHEAD		DCB  IMEDD+5
		DCB	"AHEAD"
		ALIGN	
AHEAD
		_NEST
		BL      COMPI
		DCD	BRAN-MAPOFFSET
		BL	HERE
		_DOLIT
		DCD	0
		BL	COMMA
		_UNNEST

;   REPEAT      ( A a -- )
;               Terminate a BEGIN-WHILE-REPEAT indefinite loop.

		DCD	_AHEAD-MAPOFFSET
_REPEA		DCB  IMEDD+6
		DCB	"REPEAT"
		ALIGN	
REPEA
		_NEST
		BL      AGAIN
		BL	HERE
		BL	SWAP
		BL	STORE
		_UNNEST

;   THEN        ( A -- )
;               Terminate a conditional branch structure.

		DCD	_REPEA-MAPOFFSET
_THENN		DCB  IMEDD+4
		DCB	"THEN"
		ALIGN	
THENN
		_NEST
		BL      HERE
		BL	SWAP
		BL	STORE
		_UNNEST

;   AFT         ( a -- a A )
;               Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

		DCD	_THENN-MAPOFFSET
_AFT		DCB  IMEDD+3
		DCB	"AFT"
		ALIGN	
AFT
		_NEST
		BL      DROP
		BL	AHEAD
		BL	BEGIN
		BL	SWAP
		_UNNEST

;   ELSE        ( A -- A )
;               Start the false clause in an IF-ELSE-THEN structure.

		DCD	_AFT-MAPOFFSET
_ELSEE		DCB  IMEDD+4
		DCB	"ELSE"
		ALIGN	
ELSEE
		_NEST
		BL      AHEAD
		BL	SWAP
		BL	THENN
		_UNNEST

;   WHILE       ( a -- A a )
;               Conditional branch out of a BEGIN-WHILE-REPEAT loop.

		DCD	_ELSEE-MAPOFFSET
_WHILE		DCB  IMEDD+5
		DCB	"WHILE"
		ALIGN	
WHILE
		_NEST
		BL      IFF
		BL	SWAP
		_UNNEST

;   ABORT"	( -- ; DCB> )
;               Conditional abort with an error message.

		DCD	_WHILE-MAPOFFSET
_ABRTQ		DCB  IMEDD+6
		DCB	"ABORT\""
		ALIGN	
ABRTQ
		_NEST
		BL      COMPI
		DCD	ABORQ-MAPOFFSET
		BL	STRCQ
		_UNNEST

;   _"		( -- ; DCB> )
;               Compile an inlineDCB literal.

		DCD	_ABRTQ-MAPOFFSET
_STRQ		DCB  IMEDD+2
		DCB	"$$"""
		ALIGN	
STRQ
		_NEST
		BL      COMPI
		DCD	STRQP-MAPOFFSET
		BL	STRCQ
		_UNNEST

;   ."		( -- ; DCB> )
;               Compile an inlineDCB literal to be typed out at run time.

		DCD	_STRQ-MAPOFFSET
_DOTQ		DCB  IMEDD+2
		DCB	"."""
		ALIGN	
DOTQ
		_NEST
		BL      COMPI
		DCD	DOTQP-MAPOFFSET
		BL	STRCQ
		_UNNEST

; Name compiler

;   ?UNIQUE     ( a -- a )
;               Display a warning message if the word already exists.

		DCD	_DOTQ-MAPOFFSET
_UNIQU		DCB  7
		DCB	"?UNIQUE"
		ALIGN	
UNIQU
		_NEST
		BL      DUPP
		BL	NAMEQ		;?name exists
		BL      QBRAN
		DCD	UNIQ1-MAPOFFSET		;redefinitions are OK
		BL      DOTQP
		DCB	7
		DCB	" reDef "	;but warn the user
		ALIGN	
		BL      OVER
		BL	COUNT
		BL	TYPEE		;just in case its not planned
UNIQ1          BL      DROP
		_UNNEST

;   _,n         ( na -- )
;               Build a new dictionary name using theDCB at na.

		DCD	_UNIQU-MAPOFFSET
_SNAME		DCB  3
		DCB	"$$,n"
		ALIGN	
SNAME
		_NEST
		BL      DUPP
		BL	CAT		;?null input
		BL      QBRAN
		DCD	PNAM1-MAPOFFSET
		BL      UNIQU		;?redefinition
		BL      DUPP
		BL	NAMET
		BL	CP
		BL	STORE		;save na for
 		BL      DUPP
		BL	LAST
		BL	STORE		;save na for vocabulary link
		BL      CELLM		;link address
		BL      CRRNT
		BL	AT
		BL	AT
		BL	SWAP
		BL	STORE
		_UNNEST
PNAM1          BL      STRQP
		BL	8
		DCB	" no name"	;null input
		ALIGN	
		BL      THROW

; FORTH compiler

;   _COMPILE    ( a -- )
;               Compile next word to code dictionary as a token or literal.

		DCD	_SNAME-MAPOFFSET
_SCOMP		DCB  8
		DCB	"$$COMPILE"
		ALIGN	
SCOMP
		_NEST
		BL      NAMEQ
		BL	QDUP		;defined?
		BL      QBRAN
		DCD	SCOM2-MAPOFFSET
		BL      AT
		_DOLIT
		DCD	IMEDD
		BL	ANDD		;immediate?
		BL      QBRAN
		DCD	SCOM1-MAPOFFSET
		BL      EXECU
		_UNNEST			;it's immediate, execute
SCOM1          BL      CALLC		;it's not immediate, compile
		_UNNEST		
SCOM2          BL      TNUMB
		BL	ATEXE		;try to convert to number
		BL      QBRAN
		DCD	SCOM3-MAPOFFSET
		BL      LITER
		_UNNEST			;compile number as integer
SCOM3          BL      THROW		;error

;   OVERT       ( -- )
;               Link a new word into the current vocabulary.

		DCD	_SCOMP-MAPOFFSET
_OVERT		DCB  5
		DCB	"OVERT"
		ALIGN	
OVERT
		_NEST
		BL      LAST
		BL	AT
		BL	CRRNT
		BL	AT
		BL	STORE
		_UNNEST

;   ;           ( -- )
;               Terminate a colon definition.

		DCD	_OVERT-MAPOFFSET
_SEMIS		DCB  IMEDD+COMPO+1
		DCB	";"
		ALIGN	
SEMIS
		_NEST
		_DOLIT
		_UNNEST
		BL	COMMA
		BL	LBRAC
		BL	OVERT
		_UNNEST

;   ]           ( -- )
;               Start compiling the words in the input stream.

		DCD	_SEMIS-MAPOFFSET
_RBRAC		DCB  1
		DCB	"]"
		ALIGN	
RBRAC
		_NEST
		_DOLIT
		DCD	SCOMP-MAPOFFSET
		BL	TEVAL
		BL	STORE
		_UNNEST

;   call,       ( ca -- )
;               Assemble a call instruction to ca.

		DCD	_RBRAC-MAPOFFSET
_CALLC		DCB  5
		DCB	"call,"
		ALIGN	
CALLC
		_NEST
		BL	HERE
		BL	SUBB
		BL	CELLSL
		BL	TWOM
		AND	R5,R5,#0x00FFFFFF
		ADD	R5,R5,#CALLL
		BL	COMMA		;Subroutine Threaded Code
		_UNNEST

;              ( -- ; DCB> )
;               Start a new colon definition using next word as its name.

		DCD	_CALLC-MAPOFFSET
_COLON		DCB  1
		DCB	":"
		ALIGN	
COLON
		_NEST
		BL      TOKEN
		BL	SNAME
		_DOLIT
		_NEST
		BL      COMMA
		BL	RBRAC
		_UNNEST

;   IMMEDIATE   ( -- )
;               Make the last compiled word an immediate word.

		DCD	_COLON-MAPOFFSET
_IMMED		DCB  9
		DCB	"IMMEDIATE"
		ALIGN	
IMMED
		_NEST
		_DOLIT
		DCD	IMEDD
		BL	LAST
		BL	AT
		BL	AT
		BL	ORR
		BL      LAST
		BL	AT
		BL	STORE
		_UNNEST

; Defining words

;   USER        ( u -- ; DCB> )
;               Compile a new user variable.

		DCD	_IMMED-MAPOFFSET
_USER		DCB  4
		DCB	"USER"
		ALIGN	
USER
		_NEST
		BL      TOKEN
		BL	SNAME
		BL	OVERT
		_DOLIT
		_NEST
		BL      COMMA
		_DOLIT
		DCD	DOUSE
		BL	CALLC
		BL      COMMA
		_UNNEST

;   CONSTANT        ( u -- ; DCB> )
;               Compile a new constant.

		DCD	_USER-MAPOFFSET
_CONST		DCB  8
		DCB	"CONSTANT"
		ALIGN	
CONST
		_NEST
		BL      TOKEN
		BL	SNAME
		BL	OVERT
		_DOLIT
		_NEST
		BL      COMMA
		_DOLIT
		DCD	DOCON-MAPOFFSET
		BL	CALLC
		BL      COMMA
		_UNNEST

;   CREATE      ( -- ; DCB> )
;               Compile a new array entry without allocating code space.

		DCD	_CONST-MAPOFFSET
_CREAT		DCB  6
		DCB	"CREATE"
		ALIGN	
CREAT
		_NEST
		BL      TOKEN
		BL	SNAME
		BL	OVERT
		_DOLIT
		_NEST
		BL      COMMA
		_DOLIT
		DCD	DOVAR-MAPOFFSET
		BL	CALLC
		_UNNEST

;   VARIABLE    ( -- ; DCB> )
;               Compile a new variable initialized to 0.

		DCD	_CREAT-MAPOFFSET
_VARIA		DCB  8
		DCB	"VARIABLE"
		ALIGN	
VARIA
		_NEST
		BL      CREAT
		_DOLIT
		DCD	0
		BL	COMMA
		_UNNEST

; Tools

;   _TYPE       ( b u -- )
;               Display aDCB. Filter non-printing characters.

		DCD	_VARIA-MAPOFFSET
_UTYPE		DCB  5
		DCB	"_TYPE"
		ALIGN	
UTYPE
		_NEST
		BL      TOR		;start count down loop
		B	UTYP2		;skip first pass
UTYP1          BL      DUPP
		BL	CAT
		BL	TCHAR
		BL	EMIT		;display only printable
		BL      ONEP		;increment address
UTYP2          BL      DONXT
		DCD	UTYP1-MAPOFFSET		;loop till done
		BL      DROP
		_UNNEST

;   dm+         ( a u -- a )
;               Dump u bytes from , leaving a+u on the stack.

		DCD	_UTYPE-MAPOFFSET
_DMP		DCB  3
		DCB	"dm+"
		ALIGN	
DMP
		_NEST
		BL      OVER
		_DOLIT
		DCD	4
		BL	UDOTR		;display address
		BL      SPACE
		BL	TOR		;start count down loop
		B	PDUM2		;skip first pass
PDUM1          BL      DUPP
		BL	CAT
		_DOLIT
		DCD	3
		BL	UDOTR		;display numeric data
		BL      ONEP		;increment address
PDUM2          BL      DONXT
		DCD	PDUM1-MAPOFFSET		;loop till done
		_UNNEST

;   DUMP        ( a u -- )
;               Dump u bytes from a, in a formatted manner.

		DCD	_DMP-MAPOFFSET
_DUMP		DCB  4
		DCB	"DUMP"
		ALIGN	
DUMP
		_NEST
		BL      BASE
		BL	AT
		BL	TOR
		BL	HEX		;save radix,set hex
		_DOLIT
		DCD	16
		BL	SLASH		;change count to lines
		BL      TOR
		B	DUMP4		;start count down loop
DUMP1          BL      CR
		_DOLIT
		DCD	16
		BL	DDUP
		BL	DMP		;display numeric
		BL      ROT
		BL	ROT
		BL      SPACE
		BL	SPACE
		BL	UTYPE		;display printable characters
DUMP4          BL      DONXT
		DCD	DUMP1-MAPOFFSET		;loop till done
DUMP3          BL      DROP
		BL	RFROM
		BL	BASE
		BL	STORE		;restore radix
		_UNNEST

;   .S          ( ... -- ... )
;               Display the contents of the data stack.

		DCD	_DUMP-MAPOFFSET
_DOTS		DCB  2
		DCB	".S"
		ALIGN	
DOTS
		_NEST
		BL      SPACE
		BL	DEPTH		;stack depth
		BL      TOR		;start count down loop
		B	DOTS2		;skip first pass
DOTS1          BL      RAT
		BL	PICK
		BL	DOT		;index stack, display contents
DOTS2          BL      DONXT
		DCD	DOTS1-MAPOFFSET		;loop till done
		BL      SPACE
		_UNNEST

;   !CSP        ( -- )
;               Save stack pointer in CSP for error checking.

		DCD	_DOTS-MAPOFFSET
_STCSP		DCB  4
		DCB	"!CSP"
		ALIGN	
STCSP
		_NEST
		BL      SPAT
		BL	CSP
		BL	STORE
		_UNNEST			;save pointer

;   ?CSP        ( -- )
;               Abort if stack pointer differs from that saved in CSP.

		DCD	_STCSP-MAPOFFSET
_QCSP		DCB  4
		DCB	"?CSP"
		ALIGN	
QCSP
		_NEST
		BL      SPAT
		BL	CSP
		BL	AT
		BL	XORR		;compare pointers
		BL      ABORQ
		DCB	7
		DCB	" stacks"	;abort if different
		ALIGN	2
		_UNNEST

;   >NAME       ( ca -- na | F )
;               Convert code address to a name address.

		DCD	_QCSP-MAPOFFSET
_TNAME		DCB  5
		DCB	">NAME"
		ALIGN	
TNAME
		_NEST
		BL      CNTXT
		BL	CELLM    	;ca context-2 --
TNAM1          BL      CELLP
		BL	DUPP
		BL	TOR
		BL	AT     		;ca la --
		BL	   QDUP
		BL      QBRAN
		DCD	TNAM4-MAPOFFSET		;all vocabularies checked, no match
		BL      DDUP		;ca la ca la --
TNAM2          BL      AT
		BL	DUPP		;ca la ca na na --
		BL      QBRAN
		DCD	TNAM3-MAPOFFSET		;ca la ca 0 --
		BL      DDUP
		BL	NAMET
		BL	XORR		;ca la ca na f --
		BL      QBRAN
		DCD	TNAM3-MAPOFFSET		;ca la ca na --
		BL      CELLM		;ca la ca la --
		B	TNAM2
TNAM3          BL      SWAP
		BL	DROP
		BL	SWAP
		BL	DROP		;ca 0|na --
		BL	   RFROM
		BL	SWAP 		;ca va 0|na --
		BL	   QDUP		;ca va 0|na.na --
		BL      QBRAN
		DCD	TNAM1-MAPOFFSET
		BL      SWAP
		BL	DROP
		BL	SWAP
		BL	DROP
		_UNNEST	;na
TNAM4          BL      RFROM
		BL	DDROP
		_DOLIT
		DCD	0
		_UNNEST     ;0

;   .ID         ( na -- )
;               Display the name at address.

		DCD	_TNAME-MAPOFFSET
_DOTID		DCB  3
		DCB	".ID"
		ALIGN	
DOTID
		_NEST
		BL      QDUP		;if zero no name
		BL      QBRAN
		DCD	DOTI1-MAPOFFSET
		BL      COUNT
		_DOLIT
		DCD	0x1F
		BL	ANDD		;mask lexicon bits
		BL      UTYPE
		_UNNEST			;display nameDCB
DOTI1          BL      DOTQP
		DCB	9
		DCB	" {noName}"
		ALIGN	
		_UNNEST

;   SEE         ( -- ; DCB> )
;               A simple decompiler.

		DCD	_DOTID-MAPOFFSET
_SEE		DCB  3
		DCB	"SEE"
		ALIGN	
SEE
		_NEST
		BL      TICK		;starting address
		BL      CR                
SEE1           BL      CELLP
		BL	DUPP
		BL	AT
		BL	DUPP		;?does it contain a zero
		_DOLIT
		DCD	0x800000
		BL ANDD
		BL      QBRAN
		DCD	SEE11-MAPOFFSET
		_DOLIT
		DCD	0x0FF000000
		BL ORR
		B SEE12
SEE11	
		_DOLIT
		DCD 0XFFFFFF
		BL ANDD
SEE12
		BL	TWOP
		BL	CELLS
		BL	OVER
		BL	PLUS
		BL      TNAME		;?is it a name
SEE2    BL      QDUP		;name address or zero
		BL      QBRAN
		DCD	SEE3-MAPOFFSET
		BL      SPACE
		BL	DOTID		;display name
		B	SEE4
SEE3    BL      DUPP
		BL	AT
		BL	UDOT		;display number
SEE4    BL      NUFQ		;user control
		BL      QBRAN
		DCD	SEE1-MAPOFFSET
		BL      DROP
		_UNNEST

;   WORDS       ( -- )
;               Display the names in the context vocabulary.

		DCD	_SEE-MAPOFFSET
_WORDS		DCB  5
		DCB	"WORDS"
		ALIGN	
WORDS
		_NEST
		BL      CR
		BL	CNTXT
		BL	AT		;only in context
WORS1          BL      AT
		BL	QDUP		;?at end of list
		BL      QBRAN
		DCD	WORS2-MAPOFFSET
		BL      DUPP
		BL	SPACE
		BL	DOTID		;display a name
		BL      CELLM
		B	WORS1
WORS2
          _UNNEST

; Hardware reset

;   VER         ( -- n )
;               Return the version number of this implementation.

		DCD	_WORDS-MAPOFFSET
_VERSN		DCB  3
		DCB	"VER"
		ALIGN	
VERSN
		_NEST
		_DOLIT
		DCD	VER*256+EXT
		_UNNEST

;   hi          ( -- )
;               Display the sign-on message of eForth.

		DCD	_VERSN-MAPOFFSET
_HI		DCB  2
		DCB	"hi"
		ALIGN	
HI
		_NEST
		BL	CR			;initialize I/O
		BL	DOTQP
		DCB	13
		DCB	"ARM7 eForth v"	;model
		ALIGN	
		BL	BASE
		BL	AT
		BL	HEX             ;save radix
		BL	VERSN
		BL	BDIGS
		BL	DIG
		BL	DIG
		_DOLIT
		DCD	'.'
		BL	HOLD
		BL	DIGS
		BL	EDIGS
		BL	TYPEE        	;format version number
		BL	BASE
		BL	STORE
		BL	CR
		_UNNEST      		;restore radix

;   'BOOT       ( -- a )
;               The application startup vector.

		DCD	_HI-MAPOFFSET
_TBOOT		DCB  5
		DCB	"'BOOT"
		ALIGN	
TBOOT
		_DOVAR
		DCD      HI-MAPOFFSET                      ;application to boot

;   COLD        ( -- )
;               The high level cold start sequence.

		DCD	_TBOOT-MAPOFFSET
LASTN		DCB  4
		DCB	"COLD"
		ALIGN	
COLD
		MOV R5,#0x200000		; init stacks
;		ADD R3,R5,#0xFD00  	; user area
		ADD R3,R5,#0x0200  	; user area in RAM already		  
		ADD R2,R5,#0xFD00	; return stack
		SUB	R1,R2,#0x100 	; data stack
		MOV R5,#0

		_NEST
		BL	STOIO
COLD1  	
		_DOLIT
;		DCD	UZERO-MAPOFFSET
;		_DOLIT
;		DCD	UPP
;		_DOLIT
;		DCD	ULAST-UZERO
;		BL	MOVE 		;initialize user area
		BL	PRESE		;initialize stack and TIB
		BL	TBOOT
		BL	ATEXE		;application boot
		BL	FORTH
		BL	CNTXT
		BL	AT
		BL	DUPP		;initialize search order
		BL	CRRNT
		BL	DSTOR
		BL	OVERT
		BL	QUIT		;start interpretation
		B	COLD1		;just in case
COLD2	DCD	0,0,0,0

		END
